{-# OPTIONS_GHC -fglasgow-exts #-}

module SemanticLabeling where {

import qualified List;
import qualified Show;
import qualified Product_Type;
import qualified Option;
import qualified ErrorMonad;
import qualified HOL;
import qualified Util;
import qualified Term;
import qualified Nat;
import qualified Orderings;
import qualified Trs;

data Lab a b = Lab a [b];

sshow_lab ::
  forall c d.
    (Show.Sshow c, Show.Sshow d) => SemanticLabeling.Lab c d -> [Char];
sshow_lab (SemanticLabeling.Lab f []) = Show.sshow f;
sshow_lab (SemanticLabeling.Lab f (l : ls)) =
  List.append
    (List.append (List.append (List.append (Show.sshow f) ['[']) (Show.sshow l))
      (List.concata (List.mapa (\ la -> List.append [','] (Show.sshow la)) ls)))
    [']'];

instance (Show.Sshow a,
           Show.Sshow b) => Show.Sshow (SemanticLabeling.Lab a b) where {
  sshow = SemanticLabeling.sshow_lab;
};

data ArithFun = Arg Nat.Nat | Const Nat.Nat | Sum [SemanticLabeling.ArithFun]
  | Max [SemanticLabeling.ArithFun] | Min [SemanticLabeling.ArithFun]
  | Prod [SemanticLabeling.ArithFun];

data Sl_inter a b =
  SL_Inter Nat.Nat
    [((SemanticLabeling.Lab a b, Nat.Nat), SemanticLabeling.ArithFun)];

eq_lab ::
  forall f l.
    (Eq f,
      Eq l) => SemanticLabeling.Lab f l -> SemanticLabeling.Lab f l -> Bool;
eq_lab (SemanticLabeling.Lab f list) (SemanticLabeling.Lab f' list') =
  f == f' && list == list';

instance (Eq a, Eq b) => Eq (SemanticLabeling.Lab a b) where {
  a == b = SemanticLabeling.eq_lab a b;
};

label :: forall a b. SemanticLabeling.Lab a b -> b -> SemanticLabeling.Lab a b;
label (SemanticLabeling.Lab f ls) l = SemanticLabeling.Lab f (l : ls);

unlab :: forall a b. SemanticLabeling.Lab a b -> SemanticLabeling.Lab a b;
unlab (SemanticLabeling.Lab f ls) = SemanticLabeling.Lab f (List.tl ls);

sshow_arithFun :: SemanticLabeling.ArithFun -> [Char];
sshow_arithFun (SemanticLabeling.Arg i) =
  List.append ['x', '_'] (Show.sshow_nat (Nat.Suc i));
sshow_arithFun (SemanticLabeling.Const n) = Show.sshow_nat n;
sshow_arithFun (SemanticLabeling.Sum (f : fs)) =
  List.append (SemanticLabeling.sshow_arithFun f)
    (List.concata
      (List.mapa
        (\ g -> List.append [' ', '+', ' '] (SemanticLabeling.sshow_arithFun g))
        fs));
sshow_arithFun (SemanticLabeling.Prod (f : fs)) =
  List.append
    (List.append (List.append ['('] (SemanticLabeling.sshow_arithFun f))
      (List.concata
        (List.mapa
          (\ g ->
            List.append [' ', '*', ' '] (SemanticLabeling.sshow_arithFun g))
          fs)))
    [')'];
sshow_arithFun (SemanticLabeling.Min (f : fs)) =
  List.append
    (List.append
      (List.append ['m', 'i', 'n', '('] (SemanticLabeling.sshow_arithFun f))
      (List.concata
        (List.mapa
          (\ g -> List.append [','] (SemanticLabeling.sshow_arithFun g)) fs)))
    [')'];
sshow_arithFun (SemanticLabeling.Max (f : fs)) =
  List.append
    (List.append
      (List.append ['m', 'a', 'x', '('] (SemanticLabeling.sshow_arithFun f))
      (List.concata
        (List.mapa
          (\ g -> List.append [','] (SemanticLabeling.sshow_arithFun g)) fs)))
    [')'];

sshow_sl_inter ::
  forall c d.
    (Show.Sshow c, Show.Sshow d) => SemanticLabeling.Sl_inter c d -> [Char];
sshow_sl_inter (SemanticLabeling.SL_Inter n fnas) =
  Show.showLista
    (List.append
       (List.append
         ['c', 'a', 'r', 'r', 'i', 'e', 'r', ' ', '{', '0', ',', '.', '.', '.',
           ',']
         (Show.sshow_nat n))
       ['}']
      : List.mapa
          (\ (a @ (aa, b)) ->
            let {
              (f, na) = aa;
            } in (\ ab ->
                   List.append
                     (List.append
                       (List.append
                         (List.append
                           (List.append ['['] (SemanticLabeling.sshow_lab f))
                           ['/'])
                         (Show.sshow_nat na))
                       [']', ' ', '=', ' '])
                     (SemanticLabeling.sshow_arithFun ab))
              b)
          fnas);

enum_vectors :: forall a b. [a] -> [b] -> [[(b, a)]];
enum_vectors c [] = [[]];
enum_vectors c (x : xs) =
  let {
    rec = SemanticLabeling.enum_vectors c xs;
  } in List.concata
         (List.mapa (\ vec -> List.mapa (\ ca -> (x, ca) : vec) c) rec);

assign_from_vector :: forall a b. (Eq a) => [(a, b)] -> a -> b;
assign_from_vector vec x = Option.the (Util.lookupa x vec);

eval_lab ::
  forall a b c d.
    (SemanticLabeling.Lab a b -> [c] -> c) ->
      (SemanticLabeling.Lab a b -> [c] -> b) ->
        (d -> c) ->
          Term.Term (SemanticLabeling.Lab a b) d ->
            (c, Term.Term (SemanticLabeling.Lab a b) d);
eval_lab i l alpha (Term.Var x) = (alpha x, Term.Var x);
eval_lab i l alpha (Term.Fun f ts) =
  let {
    clts = List.mapa (SemanticLabeling.eval_lab i l alpha) ts;
    cs = List.mapa fst clts;
    c = i f cs;
    lts = List.mapa snd clts;
    la = l f cs;
  } in (c, Term.Fun (SemanticLabeling.label f la) lts);

check_sl_rule_ass ::
  forall a b c d.
    (Eq a, Show.Sshow a, Eq b, Show.Sshow b, Eq c, Show.Sshow c, Eq d,
      Show.Sshow d) => Bool ->
                         (SemanticLabeling.Lab a b -> [c] -> c) ->
                           (SemanticLabeling.Lab a b -> [c] -> b) ->
                             [(Term.Term (SemanticLabeling.Lab a b) d,
                                Term.Term (SemanticLabeling.Lab a b) d)]
                               -> (d -> c) ->
                                    (Term.Term (SemanticLabeling.Lab a b) d,
                                      Term.Term (SemanticLabeling.Lab a b) d)
                                      -> Util.Eithera [Char] ();
check_sl_rule_ass mc i la lR alpha (l, r) =
  let {
    cl_ll = SemanticLabeling.eval_lab i la alpha l;
    cr_lr = SemanticLabeling.eval_lab i la alpha r;
  } in ErrorMonad.bind
         (ErrorMonad.checkB (HOL.implies mc (fst cl_ll == fst cr_lr))
           (List.append
             (List.append
               (List.append
                 (List.append
                   (List.append ['r', 'u', 'l', 'e', ' '] (Trs.showRule (l, r)))
                   [' ', 'v', 'i', 'o', 'l', 'a', 't', 'e', 's', ' ', 'm', 'o',
                     'd', 'e', 'l', ' ', 'c', 'o', 'n', 'd', 'i', 't', 'i', 'o',
                     'n', ',', ' ', '[', 'l', 'h', 's', ']', ' ', '=', ' '])
                 (Show.sshow (fst cl_ll :: c)))
               [',', ' ', '[', 'r', 'h', 's', ']', ' ', '=', ' '])
             (Show.sshow (fst cr_lr :: c))))
         (\ x ->
           ErrorMonad.checkB (List.member (snd cl_ll, snd cr_lr) lR)
             (List.append
               (List.append
                 ['l', 'a', 'b', 'e', 'l', 'e', 'd', ' ', 'r', 'u', 'l', 'e',
                   ' ']
                 (Trs.showRule (snd cl_ll, snd cr_lr)))
               [' ', 'm', 'i', 's', 's', 'i', 'n', 'g']));

check_sl_rule ::
  forall a b c d.
    (Eq a, Show.Sshow a, Eq b, Show.Sshow b, Eq c, Show.Sshow c, Eq d,
      Show.Sshow d) => (SemanticLabeling.Lab a b -> [c] -> c) ->
                         (SemanticLabeling.Lab a b -> [c] -> b) ->
                           [c] ->
                             Bool ->
                               [(Term.Term (SemanticLabeling.Lab a b) d,
                                  Term.Term (SemanticLabeling.Lab a b) d)]
                                 -> (Term.Term (SemanticLabeling.Lab a b) d,
                                      Term.Term (SemanticLabeling.Lab a b) d)
                                      -> Util.Eithera [Char] ();
check_sl_rule i la c mc lR (l, r) =
  ErrorMonad.catcha
    (ErrorMonad.checkAllMI
      (\ alpha -> SemanticLabeling.check_sl_rule_ass mc i la lR alpha (l, r))
      (List.mapa SemanticLabeling.assign_from_vector
        (SemanticLabeling.enum_vectors c
          (Util.unionI (Term.varsLI l) (Term.varsLI r)))))
    (\ x -> ErrorMonad.errora (snd x));

check_sl_model_lab_trs ::
  forall f l c v.
    (Eq f, Show.Sshow f, Eq l, Show.Sshow l, Eq c, Show.Sshow c, Eq v,
      Show.Sshow v) => (SemanticLabeling.Lab f l -> [c] -> c) ->
                         (SemanticLabeling.Lab f l -> [c] -> l) ->
                           [c] ->
                             [(Term.Term (SemanticLabeling.Lab f l) v,
                                Term.Term (SemanticLabeling.Lab f l) v)]
                               -> [(Term.Term (SemanticLabeling.Lab f l) v,
                                     Term.Term (SemanticLabeling.Lab f l) v)]
                                    -> Util.Eithera [Char] ();
check_sl_model_lab_trs i l c lR r =
  ErrorMonad.catcha
    (ErrorMonad.checkAllMI (SemanticLabeling.check_sl_rule i l c True lR) r)
    (\ x -> ErrorMonad.errora (snd x));

check_sl_lab_trs ::
  forall f l c v.
    (Eq f, Show.Sshow f, Eq l, Show.Sshow l, Eq c, Show.Sshow c, Eq v,
      Show.Sshow v) => (SemanticLabeling.Lab f l -> [c] -> c) ->
                         (SemanticLabeling.Lab f l -> [c] -> l) ->
                           [c] ->
                             [(Term.Term (SemanticLabeling.Lab f l) v,
                                Term.Term (SemanticLabeling.Lab f l) v)]
                               -> [(Term.Term (SemanticLabeling.Lab f l) v,
                                     Term.Term (SemanticLabeling.Lab f l) v)]
                                    -> Util.Eithera [Char] ();
check_sl_lab_trs i l c lP p =
  ErrorMonad.catcha
    (ErrorMonad.checkAllMI (SemanticLabeling.check_sl_rule i l c False lP) p)
    (\ x -> ErrorMonad.errora (snd x));

check_sem_lab_trs ::
  forall f l v c.
    (Show.Sshow f, Show.Sshow l,
      Show.Sshow v) => ([(Term.Term (SemanticLabeling.Lab f l) v,
                           Term.Term (SemanticLabeling.Lab f l) v)]
                         -> [(Term.Term (SemanticLabeling.Lab f l) v,
                               Term.Term (SemanticLabeling.Lab f l) v)]
                              -> Util.Eithera [Char] ())
                         -> ([(Term.Term (SemanticLabeling.Lab f l) v,
                                Term.Term (SemanticLabeling.Lab f l) v)]
                              -> [(Term.Term (SemanticLabeling.Lab f l) v,
                                    Term.Term (SemanticLabeling.Lab f l) v)]
                                   -> Util.Eithera [Char] ())
                              -> (SemanticLabeling.Lab f l -> [c] -> c) ->
                                   (SemanticLabeling.Lab f l -> [c] -> l) ->
                                     [(Term.Term (SemanticLabeling.Lab f l) v,
Term.Term (SemanticLabeling.Lab f l) v)]
                                       -> [(Term.Term (SemanticLabeling.Lab f l)
      v,
     Term.Term (SemanticLabeling.Lab f l) v)]
    -> Util.Eithera [[Char]] ();
check_sem_lab_trs check_lab check_model_lab i l r lR =
  ErrorMonad.catcha (check_model_lab lR r)
    (\ x ->
      ErrorMonad.errora
        [['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'i', 't', 'h', ' ', 'l',
           'a', 'b', 'e', 'l', 'e', 'd', ' ', 'T', 'R', 'S', ':'],
          x]);

take_default :: forall a. a -> [a] -> Nat.Nat -> a;
take_default def [] uu = def;
take_default uv (x : xs) Nat.Zero_nat = x;
take_default def (uw : xs) (Nat.Suc i) = SemanticLabeling.take_default def xs i;

eval_arithFun_unbound :: [Nat.Nat] -> SemanticLabeling.ArithFun -> Nat.Nat;
eval_arithFun_unbound nats (SemanticLabeling.Arg i) =
  SemanticLabeling.take_default Nat.Zero_nat nats i;
eval_arithFun_unbound nats (SemanticLabeling.Const n) = n;
eval_arithFun_unbound nats (SemanticLabeling.Sum []) = Nat.Zero_nat;
eval_arithFun_unbound nats (SemanticLabeling.Sum (f : fs)) =
  Nat.plus_nat (SemanticLabeling.eval_arithFun_unbound nats f)
    (SemanticLabeling.eval_arithFun_unbound nats (SemanticLabeling.Sum fs));
eval_arithFun_unbound nats (SemanticLabeling.Prod []) = Nat.one_nat;
eval_arithFun_unbound nats (SemanticLabeling.Prod (f : fs)) =
  Nat.times_nat (SemanticLabeling.eval_arithFun_unbound nats f)
    (SemanticLabeling.eval_arithFun_unbound nats (SemanticLabeling.Prod fs));
eval_arithFun_unbound nats (SemanticLabeling.Max [f]) =
  SemanticLabeling.eval_arithFun_unbound nats f;
eval_arithFun_unbound nats (SemanticLabeling.Max (f : v : va)) =
  Orderings.maxa (SemanticLabeling.eval_arithFun_unbound nats f)
    (SemanticLabeling.eval_arithFun_unbound nats
      (SemanticLabeling.Max (v : va)));
eval_arithFun_unbound nats (SemanticLabeling.Min [f]) =
  SemanticLabeling.eval_arithFun_unbound nats f;
eval_arithFun_unbound nats (SemanticLabeling.Min (f : v : va)) =
  Orderings.mina (SemanticLabeling.eval_arithFun_unbound nats f)
    (SemanticLabeling.eval_arithFun_unbound nats
      (SemanticLabeling.Min (v : va)));

eval_arithFun :: Nat.Nat -> SemanticLabeling.ArithFun -> [Nat.Nat] -> Nat.Nat;
eval_arithFun c f nats =
  Nat.mod_nat (SemanticLabeling.eval_arithFun_unbound nats f) c;

sl_inter_to_inter ::
  forall a b.
    (Eq a,
      Eq b) => SemanticLabeling.Sl_inter a b ->
                 SemanticLabeling.Lab a b -> [Nat.Nat] -> Nat.Nat;
sl_inter_to_inter (SemanticLabeling.SL_Inter c ls) fl cs =
  (case Util.lookupa (fl, List.size_list cs) ls of {
    Nothing -> Nat.Zero_nat;
    Just f -> SemanticLabeling.eval_arithFun (Nat.Suc c) f cs;
  });

get_largest_element :: forall a b. SemanticLabeling.Sl_inter a b -> Nat.Nat;
get_largest_element (SemanticLabeling.SL_Inter n uu) = n;

create_sem_lab_trs ::
  forall f l v.
    (Eq f, Show.Sshow f, Eq l, Show.Sshow l, Eq v,
      Show.Sshow v) => SemanticLabeling.Sl_inter f l ->
                         (SemanticLabeling.Lab f l -> [Nat.Nat] -> l) ->
                           [(Term.Term (SemanticLabeling.Lab f l) v,
                              Term.Term (SemanticLabeling.Lab f l) v)]
                             -> [(Term.Term (SemanticLabeling.Lab f l) v,
                                   Term.Term (SemanticLabeling.Lab f l) v)]
                                  -> Util.Eithera [[Char]] ();
create_sem_lab_trs sli l r lr =
  let {
    c = SemanticLabeling.get_largest_element sli;
    ca = List.upt Nat.Zero_nat (Nat.Suc c);
    i = SemanticLabeling.sl_inter_to_inter sli;
  } in ErrorMonad.catcha
         (SemanticLabeling.check_sem_lab_trs
           (SemanticLabeling.check_sl_lab_trs i l ca)
           (SemanticLabeling.check_sl_model_lab_trs i l ca) i l r lr)
         (\ x -> ErrorMonad.errora (SemanticLabeling.sshow_sl_inter sli : x));

map_term :: forall a b c. (a -> b) -> Term.Term a c -> Term.Term b c;
map_term fg (Term.Var x) = Term.Var x;
map_term fg (Term.Fun f ts) =
  Term.Fun (fg f) (List.mapa (SemanticLabeling.map_term fg) ts);

map_rule ::
  forall a b c.
    (a -> b) ->
      (Term.Term a c, Term.Term a c) -> (Term.Term b c, Term.Term b c);
map_rule fg lr =
  (SemanticLabeling.map_term fg (fst lr),
    SemanticLabeling.map_term fg (snd lr));

map_trsI ::
  forall a b c.
    (a -> b) ->
      [(Term.Term a c, Term.Term a c)] -> [(Term.Term b c, Term.Term b c)];
map_trsI fg r = List.mapa (SemanticLabeling.map_rule fg) r;

check_unlab_trs ::
  forall f l v.
    (Eq f, Show.Sshow f, Eq l, Show.Sshow l, Eq v,
      Show.Sshow v) => [(Term.Term (SemanticLabeling.Lab f l) v,
                          Term.Term (SemanticLabeling.Lab f l) v)]
                         -> [(Term.Term (SemanticLabeling.Lab f l) v,
                               Term.Term (SemanticLabeling.Lab f l) v)]
                              -> Util.Eithera [[Char]] ();
check_unlab_trs r uR =
  ErrorMonad.catcha
    (ErrorMonad.checkSubset (SemanticLabeling.map_trsI SemanticLabeling.unlab r)
      uR)
    (\ x ->
      ErrorMonad.errora
        [List.append (List.append ['r', 'u', 'l', 'e', ' '] (Trs.showRule x))
           [' ', 'i', 's', ' ', 'm', 'i', 's', 's', 'i', 'n', 'g', ' ', 'i',
             'n', ' ', 'u', 'n', 'l', 'a', 'b', 'e', 'l', 'e', 'd', ' ', 'T',
             'R', 'S']]);

init_lab :: forall a b. a -> SemanticLabeling.Lab a b;
init_lab f = SemanticLabeling.Lab f [];

check_sl_unlab_trs ::
  forall a b c.
    (Eq a, Eq b,
      Eq c) => [(Term.Term (SemanticLabeling.Lab a b) c,
                  Term.Term (SemanticLabeling.Lab a b) c)]
                 -> [(Term.Term (SemanticLabeling.Lab a b) c,
                       Term.Term (SemanticLabeling.Lab a b) c)]
                      -> Util.Eithera
                           (Term.Term (SemanticLabeling.Lab a b) c,
                             Term.Term (SemanticLabeling.Lab a b) c)
                           ();
check_sl_unlab_trs lR r =
  ErrorMonad.checkSubset (SemanticLabeling.map_trsI SemanticLabeling.unlab lR)
    r;

check_sem_lab_proc ::
  forall f l v c.
    (Eq f, Show.Sshow f, Eq l, Show.Sshow l, Eq v,
      Show.Sshow v) => ([(Term.Term (SemanticLabeling.Lab f l) v,
                           Term.Term (SemanticLabeling.Lab f l) v)]
                         -> [(Term.Term (SemanticLabeling.Lab f l) v,
                               Term.Term (SemanticLabeling.Lab f l) v)]
                              -> Util.Eithera [Char] ())
                         -> ([(Term.Term (SemanticLabeling.Lab f l) v,
                                Term.Term (SemanticLabeling.Lab f l) v)]
                              -> [(Term.Term (SemanticLabeling.Lab f l) v,
                                    Term.Term (SemanticLabeling.Lab f l) v)]
                                   -> Util.Eithera [Char] ())
                              -> (SemanticLabeling.Lab f l -> [c] -> c) ->
                                   (SemanticLabeling.Lab f l -> [c] -> l) ->
                                     [(Term.Term (SemanticLabeling.Lab f l) v,
Term.Term (SemanticLabeling.Lab f l) v)]
                                       -> [(Term.Term (SemanticLabeling.Lab f l)
      v,
     Term.Term (SemanticLabeling.Lab f l) v)]
    -> [(Term.Term (SemanticLabeling.Lab f l) v,
          Term.Term (SemanticLabeling.Lab f l) v)]
         -> [(Term.Term (SemanticLabeling.Lab f l) v,
               Term.Term (SemanticLabeling.Lab f l) v)]
              -> Util.Eithera [[Char]] ();
check_sem_lab_proc check_lab check_model_lab i l p r lP lR =
  ErrorMonad.bind
    (ErrorMonad.bind
      (ErrorMonad.catcha (check_lab lP p)
        (\ x ->
          ErrorMonad.errora
            [['n', 'o', 't', ' ', 'a', 'l', 'l', ' ', 'l', 'a', 'b', 'e', 'l',
               'e', 'd', ' ', 'D', 'P', 's', ' ', 'p', 'r', 'e', 's', 'e', 'n',
               't', ':'],
              x]))
      (\ x ->
        ErrorMonad.catcha (check_model_lab lR r)
          (\ xa ->
            ErrorMonad.errora
              [['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'i', 't', 'h', ' ',
                 'l', 'a', 'b', 'e', 'l', 'e', 'd', ' ', 'T', 'R', 'S', ':'],
                xa])))
    (\ x ->
      ErrorMonad.catcha (SemanticLabeling.check_sl_unlab_trs lR r)
        (\ xa ->
          ErrorMonad.errora
            [['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'i', 't', 'h', ' ',
               'l', 'a', 'b', 'e', 'l', 'e', 'd', ' ', 'T', 'R', 'S', ':'],
              List.append
                (List.append
                  ['f', 'o', 'r', ' ', 't', 'h', 'e', ' ', 'l', 'a', 'b', 'e',
                    'l', 'e', 'd', ' ', 'r', 'u', 'l', 'e', ' ']
                  (Trs.showRule xa))
                [' ', 't', 'h', 'e', 'r', 'e', ' ', 'i', 's', ' ', 'n', 'o',
                  ' ', 'u', 'n', 'l', 'a', 'b', 'e', 'l', 'e', 'd', ' ', 'c',
                  'o', 'u', 'n', 't', 'e', 'r', 'p', 'a', 'r', 't']]));

create_sem_lab_proc ::
  forall f l v.
    (Eq f, Show.Sshow f, Eq l, Show.Sshow l, Eq v,
      Show.Sshow v) => SemanticLabeling.Sl_inter f l ->
                         (SemanticLabeling.Lab f l -> [Nat.Nat] -> l) ->
                           [(Term.Term (SemanticLabeling.Lab f l) v,
                              Term.Term (SemanticLabeling.Lab f l) v)]
                             -> [(Term.Term (SemanticLabeling.Lab f l) v,
                                   Term.Term (SemanticLabeling.Lab f l) v)]
                                  -> [(Term.Term (SemanticLabeling.Lab f l) v,
Term.Term (SemanticLabeling.Lab f l) v)]
                                       -> [(Term.Term (SemanticLabeling.Lab f l)
      v,
     Term.Term (SemanticLabeling.Lab f l) v)]
    -> Util.Eithera [[Char]] ();
create_sem_lab_proc sli l r1 r2 r3 r4 =
  let {
    c = SemanticLabeling.get_largest_element sli;
    ca = List.upt Nat.Zero_nat (Nat.Suc c);
    i = SemanticLabeling.sl_inter_to_inter sli;
  } in ErrorMonad.catcha
         (SemanticLabeling.check_sem_lab_proc
           (SemanticLabeling.check_sl_lab_trs i l ca)
           (SemanticLabeling.check_sl_model_lab_trs i l ca) i l r1 r2 r3 r4)
         (\ x -> ErrorMonad.errora (SemanticLabeling.sshow_sl_inter sli : x));

}
