(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Interpretation imports RedPair Show
begin

types ('f,'c)"inter" = "'f \<Rightarrow> nat \<Rightarrow> ('c list \<Rightarrow> 'c)"

fun interpretation_of_term :: "('v \<Rightarrow> 'c) \<Rightarrow> ('f,'c) inter \<Rightarrow> ('f,'v) term \<Rightarrow> 'c" ("_\<guillemotleft>_,_\<guillemotright>" 50)
where "\<beta>\<guillemotleft>\<alpha>,Var x\<guillemotright> = \<beta> x" 
   |  "\<beta>\<guillemotleft>\<alpha>,Fun f ts\<guillemotright> = (\<alpha> f (length ts) (map (\<lambda> t. \<beta>\<guillemotleft>\<alpha>,t\<guillemotright>) ts))"

fun get_sub where "get_sub (Sub \<sigma>) = \<sigma>"

lemma interpretation_sub[simp] : "\<beta>\<guillemotleft>\<alpha>, t \<cdot> \<sigma>\<guillemotright> = (\<lambda> x. \<beta>\<guillemotleft>\<alpha>, get_sub \<sigma> x\<guillemotright>\<guillemotleft>\<alpha>, t\<guillemotright>)" 
proof (cases \<sigma>)
  case (Sub sig)
  have "\<beta>\<guillemotleft>\<alpha>, t \<cdot> Sub sig\<guillemotright> = (\<lambda> x. \<beta>\<guillemotleft>\<alpha>, sig x\<guillemotright>\<guillemotleft>\<alpha>, t\<guillemotright>)" (is "_ = (?B\<guillemotleft>\<alpha>, t\<guillemotright>)")
  proof (induct t)
    case (Var x)
    thus ?case by simp
  next
    case (Fun f ss)
    let ?ts = "map (\<lambda> s. s \<cdot> Sub sig) ss"
    have "(map (interpretation_of_term \<beta> \<alpha>) (?ts))
      = (map (interpretation_of_term ?B \<alpha>) ss)" using Fun by (induct ss, auto)
    with Fun show ?case by simp
  qed
  with Sub show ?thesis by simp
qed



locale wfcarrier =     
  fixes  carrier_s :: "'a \<Rightarrow> nat \<Rightarrow> 'a \<Rightarrow> bool" ("(_ > _ _)" [51,51,51] 50)
    and   carrier_ns :: "'a \<Rightarrow> nat \<Rightarrow> 'a \<Rightarrow> bool" ("(_ \<ge> _ _)" [51,51,51] 50)
  assumes SN: "SN {(x,y) . (x :: 'a) >n y}"
    and compat: "\<lbrakk>s \<ge>n (t :: 'a); t >n u\<rbrakk> \<Longrightarrow> s >n u"
begin

definition inter_s :: "nat \<Rightarrow> ('f,'a) inter \<Rightarrow> ('f,'v)trel"
  where "inter_s n \<alpha> \<equiv> {(s,t). (\<forall> \<beta> . (\<beta>\<guillemotleft>\<alpha>,s\<guillemotright>) >n (\<beta>\<guillemotleft>\<alpha>,t\<guillemotright>))}"

definition inter_ns :: "nat \<Rightarrow> ('f,'a) inter \<Rightarrow> ('f,'v)trel"
  where "inter_ns n \<alpha> \<equiv> {(s,t). (\<forall> \<beta> . (\<beta>\<guillemotleft>\<alpha>,s\<guillemotright>) \<ge>n (\<beta>\<guillemotleft>\<alpha>,t\<guillemotright>))}"

lemma inter_sn[simp] : 
  shows "SN (inter_s n \<alpha> :: ('f,'v)trel)" (is "SN ?P")
proof
  fix t :: "('f,'v) term"
  show "SN ?P t" 
  proof (rule ccontr)
    assume "\<not> SN ?P t"
    hence "\<exists>f . f 0 = t \<and> (\<forall>i. (f i,f(Suc i)) \<in> ?P)" by auto
    from this obtain f where "f 0 = t" and steps: "(\<forall> i. (f i, f (Suc i)) \<in> ?P)" by auto
    obtain c :: 'a where "c \<in> UNIV" by simp
    let ?b = "(\<lambda> x. c) :: ('v \<Rightarrow> 'a)"
    have "\<forall> i. (?b\<guillemotleft>\<alpha>,f i\<guillemotright>) >n (?b\<guillemotleft>\<alpha>, f (Suc i)\<guillemotright>)"
    proof 
      fix i
      from steps have "(f i, f (Suc i)) \<in> inter_s n \<alpha>" .. 
      thus "(?b\<guillemotleft>\<alpha>, f i\<guillemotright>) >n  (?b\<guillemotleft>\<alpha>, f (Suc i)\<guillemotright>)"  unfolding inter_s_def 
	by auto
    qed
    from this obtain g where nSN: "\<forall> i. ((g i) :: 'a) >n  g (Suc i)" by force
    with SN show False by (auto simp: SN_ars_def SN_elt_def)
  qed
qed
  

lemma inter_compat[simp] : "inter_s n \<alpha> O inter_ns n \<alpha> \<subseteq> inter_s n \<alpha>"
proof
  fix st
  assume "st \<in> inter_s n \<alpha> O inter_ns n \<alpha>"
  thus "st \<in> inter_s n \<alpha>" unfolding inter_s_def inter_ns_def
  proof (cases st)
    case (Pair s t)
    assume "st \<in> {(s, t). \<forall>\<beta>. (\<beta>\<guillemotleft>\<alpha>,s\<guillemotright>) >n (\<beta>\<guillemotleft>\<alpha>,t\<guillemotright>) } O
                {(s, t). \<forall>\<beta>. (\<beta>\<guillemotleft>\<alpha>,s\<guillemotright>) \<ge>n  (\<beta>\<guillemotleft>\<alpha>,t\<guillemotright>) }" (is "_ \<in> ?S O ?NS")
    with Pair obtain u where cond: "(s,u) \<in> ?NS \<and> (u,t) \<in> ?S" by auto
    have "\<forall> \<beta>. (\<beta>\<guillemotleft>\<alpha>,s\<guillemotright>) >n (\<beta>\<guillemotleft>\<alpha>,t\<guillemotright>)" 
    proof 
      fix \<beta>
      from cond have "(\<beta>\<guillemotleft>\<alpha>,u\<guillemotright>) >n (\<beta>\<guillemotleft>\<alpha>,t\<guillemotright>) " and "(\<beta>\<guillemotleft>\<alpha>,s\<guillemotright>) \<ge>n (\<beta>\<guillemotleft>\<alpha>,u\<guillemotright>) " by auto
      with compat show "(\<beta>\<guillemotleft>\<alpha>,s\<guillemotright>) >n (\<beta>\<guillemotleft>\<alpha>,t\<guillemotright>)" by blast
    qed
    with Pair show "st \<in> ?S" by simp
  qed
qed

lemma inter_s_sub[simp] : "sub_closed (inter_s n \<alpha>)" 
proof (unfold sub_closed_def inter_s_def, rule subsetI)
  fix st_sig
  assume st_sig: "st_sig \<in> CS {(s,t). \<forall>\<beta>. (\<beta>\<guillemotleft>\<alpha>,s\<guillemotright>) >n (\<beta>\<guillemotleft>\<alpha>,t\<guillemotright>) }" (is "_ \<in> CS ?P")
  thus "st_sig \<in> ?P"
  proof (cases st_sig)
    case (Pair s_sig t_sig)
    with st_sig have "(s_sig, t_sig) \<in> CS ?P" by simp
    hence "(s_sig, t_sig) \<in> ?P" 
    proof 
      fix s t \<sigma>
      assume s_sig: "s_sig = s \<cdot> \<sigma>" and tSig: "t_sig = t \<cdot> \<sigma>" and st: "(s,t) \<in> ?P"
      thus "(s_sig, t_sig) \<in> ?P" by auto
    qed
    with Pair show ?thesis by simp
  qed
qed

lemma inter_ns_sub[simp] : "sub_closed (inter_ns n \<alpha>)" 
proof (unfold sub_closed_def inter_ns_def, rule subsetI)
  fix st_sig
  assume st_sig: "st_sig \<in> CS {(s,t). \<forall>\<beta>. (\<beta>\<guillemotleft>\<alpha>,s\<guillemotright>) \<ge>n  (\<beta>\<guillemotleft>\<alpha>,t\<guillemotright>) }" (is "_ \<in> CS ?P")
  thus "st_sig \<in> ?P"
  proof (cases st_sig)
    case (Pair s_sig t_sig)
    with st_sig have "(s_sig, t_sig) \<in> CS ?P" by simp
    hence "(s_sig, t_sig) \<in> ?P" 
    proof 
      fix s t \<sigma>
      assume s_sig: "s_sig = s \<cdot> \<sigma>" and t_sig: "t_sig = t \<cdot> \<sigma>" and st: "(s,t) \<in> ?P"
      thus "(s_sig, t_sig) \<in> ?P" by auto
    qed
    with Pair show ?thesis by simp
  qed
qed

end



types ('f,'c,'d) "polyinter" = "'f \<Rightarrow> nat \<Rightarrow> ('c list) \<times> 'd"

locale polyinter = wfcarrier + 
  fixes plus :: "'a \<Rightarrow> 'a \<Rightarrow> 'a" (* nat vec + nat vec => nat vec *)
  and   zero :: "nat \<Rightarrow> 'a"            (* nat vec zero *)
  and   add :: "'a \<Rightarrow> 'd \<Rightarrow> 'a"  (* nat vec + int vec => nat vec (monus) *)
  and   mul_a :: "'b \<Rightarrow> 'a \<Rightarrow> 'a" (* nat mat * nat vec => nat vec *)
  and   wf_a :: "nat \<Rightarrow> 'a \<Rightarrow> bool"
  and   wf_b :: "nat \<Rightarrow> 'b \<Rightarrow> bool"
  and   wf_d :: "nat \<Rightarrow> 'd \<Rightarrow> bool"
  assumes add_mono: "\<lbrakk>wf_d n d; x \<ge>n y\<rbrakk> \<Longrightarrow> add x d \<ge>n add y d"
  and     plus_mono: "\<lbrakk>x \<ge>n y; x' \<ge>n y'\<rbrakk> \<Longrightarrow> plus x x'  \<ge>n plus y y'"
  and     mul_a_mono: "\<lbrakk>wf_b n b; x \<ge>n y\<rbrakk> \<Longrightarrow> mul_a b x \<ge>n mul_a b y"
  and     geq_refl: "wf_a n x \<Longrightarrow> (x :: 'a) \<ge>n x"

(* now let us define some carrier, first take the naturals *)


interpretation nat_wfcarrier: wfcarrier "(\<lambda> x n y. x > y) :: nat \<Rightarrow> nat \<Rightarrow> nat \<Rightarrow> bool" "(\<lambda> x n y. x \<ge> y)"
proof (unfold_locales, auto)
  have "{(x,y). (x :: nat) < y} = ({(x,y). (x :: nat) > y}^-1)" by auto
  with wf_less have "wf ({(x,y). (x :: nat) > y}^-1)" by simp 
  thus "SN {(x,y). (x :: nat) > y}" by (rule wf_imp_SN)
qed

(* and implement polynomials with negative constants *)
datatype cint = Neg nat | Pos nat

fun add :: "nat \<Rightarrow> cint \<Rightarrow> nat" 
where "add y (Neg x)  = y - x"
  |   "add y (Pos x)  = y + x"

lemma add_monotone : "n \<ge> m \<Longrightarrow> add n c \<ge> add m c"
by (cases c, auto)


(* polynomials with negative constants from a polynomial interpretation *)
interpretation nat_poly_inter : polyinter "(\<lambda> x n y. x > y) :: nat \<Rightarrow> nat \<Rightarrow> nat \<Rightarrow> bool" "\<lambda> x n y. x \<ge> y" "op +" "\<lambda> x. 0" "add" "op *" "\<lambda> n x. True" "\<lambda> n x. True" "\<lambda> n x. True"
by (unfold_locales, auto simp: add_monotone)

(* the part below has to be adapted to care about the well-formed predicates ! *)

(*

context polyinter
begin

fun pol :: "('c,'b,'d)polyinter \<Rightarrow> ('c,'a) inter"
where "pol \<alpha> f i = (\<lambda> cs. (let (args,c) = \<alpha> f i in add (foldr (\<lambda> x s. (plus (mulA (fst x) (snd x)) s)) (zip args cs) zero) c)) " 


lemma interCC[simp] : fixes \<alpha> :: "('c,'b,'d)polyinter" 
  shows "one_step_ctxt_closed ((interNS (pol \<alpha>))  :: ('c,'v)trel)" 
proof (unfold one_step_ctxt_closed_def interNS_def, (rule allI)+, rule impI)
  fix f :: 'c
    and bef :: "('c,'v) term list" 
    and s :: "('c,'v) term" 
    and t :: "('c,'v) term" 
    and aft :: "('c,'v) term list" 
  let ?P = "{(s,t). \<forall> \<beta>. (\<beta>\<guillemotleft>pol \<alpha>,s\<guillemotright>) \<ge> (\<beta>\<guillemotleft>pol \<alpha>,t\<guillemotright>)}"
  assume "(s, t) \<in> ?P"
  hence st: "\<forall> \<beta>. (\<beta>\<guillemotleft>pol \<alpha>, s\<guillemotright>) \<ge> (\<beta>\<guillemotleft>pol \<alpha>, t\<guillemotright>)" by (simp)
  show "(Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> ?P"
  proof (cases "\<alpha> f ((length bef) + 1 + (length aft))")
    case (Pair list c)
    let ?lsbb = "\<lambda> list s bef \<beta>. (foldr (\<lambda> x su. (plus (mulA (fst x) (snd x)) su)) (zip list
                 (map (interpretationOfTerm \<beta> (pol \<alpha>)) bef @
                  (\<beta>\<guillemotleft>pol \<alpha>,s\<guillemotright>) #
                  map (interpretationOfTerm \<beta> (pol \<alpha>)) aft)) zero)"
    have "\<forall> \<beta> . ?lsbb list s bef \<beta> \<ge> ?lsbb list t bef \<beta>" 
    proof
      fix \<beta>
      from st have stb: "(\<beta>\<guillemotleft>pol \<alpha>,s\<guillemotright>) \<ge> (\<beta>\<guillemotleft>pol \<alpha>,t\<guillemotright>)" .. 
      show "?lsbb list s bef \<beta> \<ge> ?lsbb list t bef \<beta>"
      proof (induct bef arbitrary: list)
	case Nil
	thus ?case using stb 
	proof (cases list, simp add: geqRefl)
	  case (Cons n ns)
	  from stb mulAMono have stbn: "mulA n (\<beta>\<guillemotleft>pol \<alpha>,s\<guillemotright>) \<ge> mulA n (\<beta>\<guillemotleft>pol \<alpha>,t\<guillemotright>)" by simp
	  from Cons have "?lsbb list s [] \<beta> = plus (mulA n (\<beta>\<guillemotleft>pol \<alpha>, s\<guillemotright>)) (foldr (\<lambda> x su. (plus (mulA (fst x) (snd x)) su)) (zip ns (map (interpretationOfTerm \<beta> (pol \<alpha>)) aft)) zero)" 
	    and "?lsbb list t [] \<beta> = plus (mulA n (\<beta>\<guillemotleft>pol \<alpha>, t\<guillemotright>)) (foldr (\<lambda> x su. (plus (mulA (fst x) (snd x)) su)) (zip ns (map (interpretationOfTerm \<beta> (pol \<alpha>)) aft)) zero)" by auto
	  with stbn plusMono geqRefl show ?thesis by arith
	qed
      next
	case (Cons b befs)
	thus ?case by (cases list, simp add: geqRefl, simp add: plusMono geqRefl)
      qed
    qed
    with Pair show ?thesis by (auto simp: addMono)
  qed
qed
end



(* next we need a check to test whether two terms are in relation,
   here we have to start to build polynomials  *)

datatype ('v,'c,'d) poly = Poly "('c \<times> 'v) list" 'd (* currently only linear polys *)


instantiation poly :: (sshow,sshow,sshow) sshow
begin
fun show_pair :: "('b \<times> 'v :: sshow) \<Rightarrow> sstring" 
where "show_pair (b,v) = (sshow b +++ ''*'' +++ sshow v)"
fun show_entries :: "('a,'b,'c)poly \<Rightarrow> sstring list"
where "show_entries (Poly vp c) = (filter (\<lambda> entry. entry \<noteq> []) (sshow c # map show_pair vp))"
fun merge_entries :: "sstring list \<Rightarrow> sstring"
where "merge_entries [] = []"
    | "merge_entries (x # xs) = ('' + '' +++ x +++ (merge_entries xs))"
fun sshow_poly :: "('a,'b,'c)poly \<Rightarrow> sstring"
 where "sshow_poly p = (case  show_entries p of Nil \<Rightarrow> ''0''
                             |      (Cons d es) \<Rightarrow> (d +++ (merge_entries es)))"
instance ..
end


locale polyinterImpl = polyinter + 
  fixes addD :: "'d :: sshow \<Rightarrow> 'd \<Rightarrow> 'd" (* int vec + int vec => int vec *)
  and   zeroD :: "'d"           (* int vec 0 *)
  and   mulD :: "'b :: sshow \<Rightarrow> 'd \<Rightarrow> 'd" (* nat mat * int vec => int vec *)
  and   mulB :: "'b \<Rightarrow> 'b \<Rightarrow> 'b" (* nat mat * nat mat => nat mat *)
  and   addB :: "'b \<Rightarrow> 'b \<Rightarrow> 'b" (* nat mat + nat mat => nat mat *)
  and   oneB :: "'b" (* nat mat 1 *)
  and   maxZero :: "'d \<Rightarrow> 'd" (* int vec \<Rightarrow> int vec (where every entry is made positive) *)
  and   zeroB :: "'b"           (* nat mat 0 *)
  and   geqB :: "'b \<Rightarrow> 'b \<Rightarrow> bool" (* \<ge> on nat matrices *)
  and   geqD :: "'d \<Rightarrow> 'd \<Rightarrow> bool" (* \<ge> on int vec *)
  and   gtD :: "'d \<Rightarrow> 'd \<Rightarrow> bool" (* > on int vec *)
  assumes plusComm: "plus n m = plus m n"
  and     plusAssoc[simp]: "plus (plus n m) k = plus n (plus m k)"
  and     mulDistrib: "(mulA :: 'b \<Rightarrow> 'a \<Rightarrow> 'a) l (plus n m) = plus (mulA l n) (mulA l m)"
  and     mulAZero[simp]: "mulA b zero = zero"
  and     plusZero[simp]: "plus a zero = a"
  and     mulAmulB[simp]: "mulA (mulB b c) a = mulA b (mulA c a)"
  and     mulAadd: "mulA b (add a d) = add (mulA b a) (mulD b d)"
  and     mulAaddB: "mulA (addB b1 b2) a = plus (mulA b1 a) (mulA b2 a)"
  and     maxZeroAddZero: "add (zero :: 'a) d \<ge> add zero (maxZero d)"
  and     maxZeroAdd: "add a (maxZero d) \<ge> add (a :: 'a) d"
  and     plusAdd:  "plus (add (a :: 'a) d) (add a' d') \<ge> add (plus a a') (addD d d')"
  and     addZeroD[simp]: "add a zeroD = a"
  and     mulAoneB[simp]: "mulA oneB a = a"
  and     mulAzeroB[simp]: "mulA zeroB a = zero"
  and     geqTrans: "\<lbrakk>a \<ge> (a' :: 'a); a' \<ge> a''\<rbrakk> \<Longrightarrow> a \<ge> a''" 
  and     geqGt: "\<lbrakk>a \<ge> (a' :: 'a); a' > a''; a'' \<ge> a'''\<rbrakk> \<Longrightarrow> a > a'''" 
  and     addAddD: "add (add (a :: 'a) (d :: 'd)) (d' :: 'd) \<ge> add a (addD d d')"
  and     posZeroD: "maxZero zeroD = zeroD"
  and     posMulD: "maxZero d = d \<Longrightarrow> maxZero (mulD b d) = mulD b d"
  and     posAddD: "\<lbrakk>maxZero d = d; maxZero d' = d'\<rbrakk> \<Longrightarrow> maxZero (addD d d') = addD d d'"
  and     posMaxZero: "maxZero (maxZero d) = maxZero d"
  and     posPlusAdd:  "\<lbrakk>maxZero d = d; maxZero d' = d'\<rbrakk> \<Longrightarrow> add (plus a a') (addD d d') = plus (add (a :: 'a) d) (add a' d')"
  and     posAddAddD: "maxZero d = d \<Longrightarrow> add a (addD d d') = add (add a d) d'"
  and     geqBMono: "geqB (b1 :: 'b) b2 \<Longrightarrow> (mulA b1 (a :: 'a)) \<ge> (mulA b2 a)"
  and     geqDMono: "geqD d d' \<Longrightarrow> add (a :: 'a) d \<ge> add a d'"
  and     gtDMono:  "gtD  d d' \<Longrightarrow> add (a :: 'a) d > add a d'"
  and     zeroLeast: "(a :: 'a) \<ge> zero"
  and     mulAMonoLeft: "geqB (b1 :: 'b) b2 \<Longrightarrow> (mulA b1 (a :: 'a) \<ge> mulA b2 a)"
begin

lemma plusZeroL[simp] : "plus zero a = a" using plusZero plusComm by auto


abbreviation polyZero :: "('v,'b,'d)poly" where "polyZero \<equiv> Poly [] zeroD"


fun lookupVar :: "'v \<Rightarrow> ('b \<times> 'v) list \<Rightarrow> ('b \<times> ('b \<times> 'v) list) option"
  where "lookupVar x [] = None"
    |   "lookupVar x ((c,y) # vp) = (if x = y then Some (c, vp) else 
             (case lookupVar x vp of None \<Rightarrow> None | Some (d,vpp) \<Rightarrow> Some (d, (c,y) # vpp)))"

fun sumPolyVars :: "('b \<times> 'v)list \<Rightarrow> ('b \<times> 'v)list \<Rightarrow> ('b \<times> 'v)list"
  where "sumPolyVars [] vp2 = vp2"
     |  "sumPolyVars ((c,x) # vp1) vp2 = (case lookupVar x vp2 of 
                  None \<Rightarrow> (c,x) # sumPolyVars vp1 vp2 
                | Some (d, vpp2) \<Rightarrow> (addB c d, x) # sumPolyVars vp1 vpp2)" 

fun cleanPolyVars :: "('b \<times> 'v) list \<Rightarrow> ('b \<times> 'v) list"
where "cleanPolyVars [] = []" 
   |  "cleanPolyVars ((b,x) # vp ) = ((if b = zeroB then [] else [(b,x)]) @ cleanPolyVars vp)"


fun polySum :: "('v,'b,'d) poly \<Rightarrow> ('v,'b,'d) poly \<Rightarrow> ('v,'b,'d) poly" (infix "+" 65)
  where "(Poly vp1 c) + (Poly vp2 d) = Poly (cleanPolyVars (sumPolyVars vp1 vp2)) (addD c d)"

fun polyMult :: "'b \<Rightarrow> ('v,'b,'d)poly \<Rightarrow> ('v,'b,'d)poly" (infix "*" 70)
  where "n * (Poly vp c) = (Poly (map (\<lambda> bx. (mulB n (fst bx), snd bx)) vp) (mulD n c))"


fun polyEvalVP :: "('v \<Rightarrow> 'a) \<Rightarrow> ('b \<times> 'v) list \<Rightarrow> 'a" 
  where "polyEvalVP \<beta> vp = (foldr (\<lambda> bv su. plus (mulA (fst bv) (\<beta> (snd bv))) su) vp zero)"

lemma polyEvalVPEmpty[simp] : "polyEvalVP \<beta> [] = zero" by simp

lemma cleanPolyVarsSound[simp]: "polyEvalVP \<beta> (cleanPolyVars vp) = polyEvalVP \<beta> vp"
proof (induct vp, simp)
  case (Cons bx vp)
  thus ?case by (cases bx, auto simp:plusComm[where n = zero])
qed

declare cleanPolyVars.simps[simp del]

fun polyEval :: "('v \<Rightarrow> 'a ) \<Rightarrow> ('v,'b,'d) poly \<Rightarrow> 'a" 
  where "polyEval \<beta> (Poly vp c) = add (polyEvalVP \<beta> vp) c"

fun polyC :: "('v,'b,'d)poly \<Rightarrow> 'd" where "polyC (Poly _ c) = c"
fun polyVP :: "('v,'b,'d)poly \<Rightarrow> ('b \<times> 'v)list" where "polyVP (Poly vp _) = vp"


lemma polyMultSound[simp] : "polyEval \<beta> (b * p) = mulA b (polyEval \<beta> p)"
proof (cases p)
  case (Poly vp d)
  have "polyEval \<beta> (b * Poly vp d) = mulA b (polyEval \<beta> (Poly vp d))" 
  proof -
    let ?left = "foldr (\<lambda> bv. plus (mulA (fst bv) (\<beta> (snd bv)))) (map (\<lambda> bx. (mulB b (fst bx), snd bx)) vp) zero"
    let ?right = "foldr (\<lambda>bv. plus (mulA (fst bv) (\<beta> (snd bv)))) vp zero"
    have lmr: "?left = mulA b ?right" 
    proof (induct vp, simp)
      case (Cons nv vpp)
      thus ?case 
      proof (cases nv)
	case (Pair n v)
	with Cons show ?thesis by (auto simp: mulDistrib)
      qed
    qed
    show ?thesis by (simp add:lmr, simp add:mulAadd)
  qed
  with Poly show ?thesis by simp
qed


lemma sumPolyVarsSound[simp] : "polyEvalVP \<beta> (sumPolyVars vp vq) = plus (polyEvalVP \<beta> vp) (polyEvalVP \<beta> vq)"
proof simp
  show "foldr (\<lambda>bv. plus (mulA (fst bv) (\<beta> (snd bv)))) (sumPolyVars vp vq) zero =
    plus (foldr (\<lambda>bv. plus (mulA (fst bv) (\<beta> (snd bv)))) vp zero) (foldr (\<lambda>bv. plus (mulA (fst bv) (\<beta> (snd bv)))) vq zero)"
  proof (induct vp arbitrary: vq)
    case Nil thus ?case by (simp add: plusComm[where n = zero])
  next
    case (Cons nv vpp)
    show ?case
    proof (cases nv)
      case (Pair b x)
      have "foldr (\<lambda>bv. plus (mulA (fst bv) (\<beta> (snd bv)))) (sumPolyVars ((b,x) # vpp) vq) zero =
          plus (foldr (\<lambda>bv. plus (mulA (fst bv) (\<beta> (snd bv)))) ((b,x) # vpp) zero) (foldr (\<lambda>bv. plus (mulA (fst bv) (\<beta> (snd bv)))) vq zero)"
      proof (cases "lookupVar x vq")
	case None
	with Cons show ?thesis by simp
      next
	case (Some evq)
	show ?thesis 
	proof (cases evq)
	  case (Pair e vqq)
	  with Some have lookup: "lookupVar x vq = Some (e, vqq)" by simp
	  from lookup have "plus (mulA e (\<beta> x)) (foldr (\<lambda>bv. plus (mulA (fst bv) (\<beta> (snd bv)))) vqq zero) = 
	      foldr (\<lambda>bv. plus (mulA (fst bv) (\<beta> (snd bv)))) vq zero" (is "plus ?e ?vqq = ?vq") 
	  proof (induct vq arbitrary: e vqq, simp)
	    case (Cons cy vp)
	    show ?case 
	    proof (cases cy)
	      case (Pair c y)
	      with Cons have lookup2: "lookupVar x ((c,y) # vp) = Some (e, vqq)" by simp
	      hence "plus (mulA e (\<beta> x)) (foldr (\<lambda>bv. plus (mulA (fst bv) (\<beta> (snd bv)))) vqq zero) = 
		foldr (\<lambda>bv. plus (mulA (fst bv) (\<beta> (snd bv)))) ((c,y) # vp) zero" 
	      proof (cases "x=y", simp, cases "lookupVar x vp", simp)
		fix dvpp
		assume xy: "x \<noteq> y" and lookup3: "lookupVar x vp = Some dvpp"
		show ?thesis
		proof (cases dvpp)
		  case (Pair d vpp)
		  have swap: "\<And> g. plus (mulA e (\<beta> x)) (plus (mulA c (\<beta> y)) g) = plus (mulA c (\<beta> y)) (plus (mulA e (\<beta> x)) g)" using plusAssoc plusComm by force
		  from Pair lookup3 have lookup4: "lookupVar x vp = Some (d, vpp)" by simp
		  with lookup2 Cons xy swap show ?thesis by auto
		qed
	      qed
	      with Pair show ?thesis by simp
	    qed 
	  qed
	  let ?vpp = "foldr (\<lambda>bv. plus (mulA (fst bv) (\<beta> (snd bv)))) vpp zero"
	  let ?b = "mulA b (\<beta> x)"
	  have "plus (mulA (addB b e) (\<beta> x)) (plus ?vpp ?vqq) = plus ?b (plus ?e (plus ?vpp ?vqq))" using mulAaddB by simp 
	  also have "\<dots> = plus (plus ?e ?vqq) (plus ?b ?vpp)" using plusAssoc plusComm by force
	  also have "\<dots> = plus ?vq (plus ?b ?vpp)" using `plus ?e ?vqq = ?vq` by simp
	  also have "\<dots> = plus ?b (plus ?vpp ?vq)" using plusComm plusAssoc by force
	  finally show ?thesis using Cons lookup by simp	    
	qed
      qed			    
      with Pair show ?thesis by simp
    qed
  qed
qed

declare sumPolyVars.simps polyEvalVP.simps[simp del]




function Pleft :: "('c,'b,'d)polyinter \<Rightarrow> ('c,'v) term \<Rightarrow> ('v,'b,'d) poly" 
where "Pleft \<alpha> (Var x) = (Poly [(oneB,x)] zeroD)" 
   |  "Pleft \<alpha> (Fun f ts) = (let (list, c) = \<alpha> f (length ts) in 
              (case (foldr (\<lambda> ct . op + ((fst ct) *  (Pleft \<alpha> (snd ct)))) (zip list ts) polyZero) + Poly [] c of
                  Poly [] d \<Rightarrow> Poly [] (maxZero d) | p \<Rightarrow> p))"
by (pat_completeness, auto)

termination Pleft
proof
  fix \<alpha> f and ts :: "('c,'v) term list" and cl const list ct
  assume "ct \<in> set (zip list ts)"
  hence "ct mem (zip list ts)" by (auto simp: mem_iff)
  hence "snd ct mem ts" proof (induct ts arbitrary: list, simp)
    case (Cons a ts)
    show ?case 
    proof (cases list)
      case Nil with Cons show ?thesis by simp
    next
      case (Cons b ll)
      from prems show ?thesis by (cases "ct = (b,a)", auto)
    qed
  qed
  thus "((\<alpha>, snd ct), (\<alpha>, Fun f ts)) \<in> measures [size \<circ> snd]" by (auto simp: mem_iff)
next
  show "wf (measures [size \<circ> snd])" by simp
qed


lemma polySumSound1 : "plus (polyEval \<beta> p)  (polyEval \<beta> q) \<ge> polyEval \<beta> (polySum p q)"
proof (cases p)
  case (Poly vp c)
  show ?thesis
  proof (cases q)
    case (Poly vq d)
    from prems show ?thesis by (simp add: plusAdd)
  qed
qed


lemma PleftSound : "(\<beta>\<guillemotleft>pol \<alpha>, t\<guillemotright>) \<ge> polyEval \<beta> (Pleft \<alpha> t)" 
proof (induct t, simp add: polyEvalVP.simps geqRefl)
  case (Fun f ts)
  show ?case 
  proof (cases "\<alpha> f (length ts)")
    case (Pair vp c)
    with Fun show ?thesis proof simp     
      assume "\<And>s. s \<in> set ts \<Longrightarrow>  (\<beta>\<guillemotleft>\<lambda>a b c. let (args, y) = \<alpha> a b in add (foldr (\<lambda> x. plus (mulA (fst x) (snd x))) (zip args c) zero) y,s\<guillemotright>) \<ge> polyEval \<beta> (Pleft \<alpha> s)" 
      hence "foldr (\<lambda>x. plus (mulA (fst x) (snd x)))
        (zip vp (map (interpretationOfTerm \<beta> (\<lambda>a b c. let (args, y) = \<alpha> a b
        in add (foldr (\<lambda>x. plus (mulA (fst x) (snd x))) (zip args c) zero) y)) ts)) zero \<ge> 
	polyEval \<beta> (foldr (\<lambda> ct. op + (fst ct * Pleft \<alpha> (snd ct))) (zip vp ts) polyZero)" (is "?PInterVP \<ge> polyEval \<beta> ?PLeftVP")
      proof (induct vp arbitrary:ts)
	case Nil
	thus ?case by (cases ts, auto simp:polyEvalVP.simps geqRefl)
      next
	case (Cons n vpp)
	thus ?case proof (cases ts, simp add:polyEvalVP.simps geqRefl)
	  case (Cons t tss)
	  from prems have ind1: "(\<beta>\<guillemotleft>\<lambda>a b c. let (args, y) = \<alpha> a b in add (foldr (\<lambda> x. plus (mulA (fst x) (snd x))) (zip args c) zero) y,t\<guillemotright>) \<ge> polyEval \<beta> (Pleft \<alpha> t)" by simp
	  from prems have ind2: 
	    "foldr (\<lambda>x. plus (mulA (fst x) (snd x))) (zip vpp (map (interpretationOfTerm \<beta> (\<lambda>a b c. let (args, a) = \<alpha> a b
            in add (foldr (\<lambda>x. plus (mulA (fst x) (snd x))) (zip args c) zero) a)) tss)) zero \<ge>
	    polyEval \<beta>  (foldr (\<lambda>ct. op + (fst ct * Pleft \<alpha> (snd ct))) (zip vpp tss) polyZero)" by simp
	  let ?exp1 = "polyEval \<beta> (n * Pleft \<alpha> t + (foldr (\<lambda>ct. op + (fst ct * Pleft \<alpha> (snd ct))) (zip (vpp) (tss)) polyZero))"
	  let ?exp2 = "plus (polyEval \<beta> (n * Pleft \<alpha> t)) (polyEval \<beta> (foldr (\<lambda>ct. op + (fst ct * Pleft \<alpha> (snd ct))) (zip (vpp) (tss)) polyZero))"
	  let ?exp3 = "(foldr (\<lambda>x. plus (mulA (fst x) (snd x))) (zip (n # vpp)
            (map (interpretationOfTerm \<beta> (\<lambda>a b c. let (args, y) = \<alpha> a b in add (foldr (\<lambda>x. plus (mulA (fst x) (snd x)))
            (zip args c) zero) y)) (t # tss))) zero)" 
	  have "?exp2 \<ge> ?exp1" by (rule polySumSound1)
	  have "?exp3 \<ge> ?exp2" by (simp, rule plusMono, rule mulAMono, simp add: ind1, simp add: ind2)
	  have res: "?exp3 \<ge> ?exp1" using geqTrans `?exp2 \<ge> ?exp1` `?exp3 \<ge> ?exp2` by blast
	  from Cons show ?thesis using res by simp
	qed
      qed
      from this obtain PInterVP PLeftVP where eq: "PInterVP = ?PInterVP \<and> PLeftVP = ?PLeftVP" and "PInterVP \<ge> polyEval \<beta> PLeftVP" by auto
      have "add PInterVP c \<ge> polyEval \<beta> (case PLeftVP + Poly [] c of
         Poly [] d \<Rightarrow> Poly [] (maxZero d)
         | Poly (a # lista) d \<Rightarrow> Poly (a # lista) d)"
      proof (cases "PLeftVP")
	case (Poly vpp d)
	with `PInterVP \<ge> polyEval \<beta> PLeftVP`
	have part0: "PInterVP \<ge> polyEval \<beta> (Poly vpp d)" by simp
	have part1: "add PInterVP c \<ge> polyEval \<beta> (Poly vpp d + Poly [] c)" by (simp, rule geqTrans, rule addMono, rule part0, auto simp: polyEvalVP.simps addAddD)
	have "add PInterVP c \<ge> polyEval \<beta> (case (Poly vpp d + Poly [] c) of Poly [] e \<Rightarrow> Poly [] (maxZero e) | Poly (a # lista) e \<Rightarrow> Poly (a # lista) e)" 
	proof (cases "polyVP (Poly vpp d + Poly [] c)")
	  case (Cons a ap) 
	  thus ?thesis using part1 by simp
	next
	  case Nil
	  hence nearlyDone: "add PInterVP c \<ge> add zero (addD d c)" using part1 by (simp add: polyEvalVP.simps)
	  have "add PInterVP c \<ge> add zero (maxZero (addD d c))" 
	  proof (rule geqTrans)
	    from nearlyDone show "add PInterVP c \<ge> add zero (addD d c)" .
	  next
	    show "add zero (addD d c) \<ge> add zero (maxZero (addD d c))" by (rule maxZeroAddZero)
	  qed
	  thus ?thesis using Nil by (simp add: polyEvalVP.simps)
	qed
	with Poly show ?thesis by simp
      qed
      with eq show "add ?PInterVP c \<ge> polyEval \<beta> (case ?PLeftVP + Poly [] c of
         Poly [] d \<Rightarrow> Poly [] (maxZero d)
         | Poly (a # lista) d \<Rightarrow> Poly (a # lista) d)" by simp
    qed
  qed  
qed


function Pright :: "('c,'b,'d)polyinter \<Rightarrow> ('c,'v) term \<Rightarrow> ('v,'b,'d) poly" 
where "Pright \<alpha> (Var x) = (Poly [(oneB,x)] zeroD)" 
   |  "Pright \<alpha> (Fun f ts) = (let (list, c) = \<alpha> f (length ts) in 
              (case (foldr (\<lambda> ct . op + ((fst ct) *  (Pright \<alpha> (snd ct)))) (zip list ts) polyZero) + Poly [] c of
                  Poly vp d \<Rightarrow> Poly vp (maxZero d)))"
by (pat_completeness, auto)

termination Pright
proof
  fix \<alpha> f and ts :: "('f,'v) term list" and cl const list ct
  assume "ct \<in> set (zip list ts)"
  hence "ct mem (zip list ts)" by (auto simp: mem_iff)
  hence "snd ct mem ts" proof (induct ts arbitrary: list, simp)
    case (Cons a ts)
    show ?case 
    proof (cases list)
      case Nil with Cons show ?thesis by simp
    next
      case (Cons b ll)
      from prems show ?thesis by (cases "ct = (b,a)", auto)
    qed
  qed
  thus "((\<alpha>, snd ct), (\<alpha>, Fun f ts)) \<in> measures [size \<circ> snd]" by (auto simp: mem_iff)
next
  show "wf (measures [size \<circ> snd])" by simp
qed

abbreviation isPositive :: "'d \<Rightarrow> bool"
where "isPositive x \<equiv> (maxZero x = x)"


lemma sumPositive : assumes "isPositive (polyC p)" and "isPositive(polyC q)"
  shows "isPositive(polyC (p+q))"
proof (cases p)
  case (Poly vp c)
  show ?thesis
  proof (cases q)
    case (Poly vq d)
    from prems show ?thesis by (simp add: posAddD)
  qed
qed


lemma PrightPositive : "isPositive(polyC (Pright \<alpha> t))"
proof (cases t, simp add: posZeroD)
  case (Fun f ts)
  thus ?thesis 
  proof (cases "\<alpha> f (length ts)")
    case (Pair list c)
    with Fun show ?thesis
    proof (cases "foldr (\<lambda> ct. op + (fst ct *  Pright \<alpha> (snd ct))) (zip list ts) polyZero + Poly [] c") 
      case (Poly vp d)
      with Fun Pair show ?thesis by (simp add: posMaxZero)
    qed
  qed
qed


lemma polySumSound2 : assumes 
  "isPositive(polyC p)" and "isPositive(polyC q)"
  shows "polyEval \<beta> (polySum p q) = plus (polyEval \<beta> p) (polyEval \<beta> q)"
proof (cases p)
  case (Poly vp c)
  show ?thesis
  proof (cases q)
    case (Poly vq d)
    from prems show ?thesis 
      by (simp add: posPlusAdd)
  qed
qed

lemma PrightSound : "polyEval \<beta> (Pright \<alpha> t) \<ge> (\<beta>\<guillemotleft>pol \<alpha>, t\<guillemotright>)"
proof (induct t, simp add: polyEvalVP.simps geqRefl)
  case (Fun f ts)
  show ?case 
  proof (cases "\<alpha> f (length ts)")
    case (Pair vp c)
    with Fun show ?thesis proof simp
      assume "\<And>s. s \<in> set ts \<Longrightarrow> polyEval \<beta> (Pright \<alpha> s) \<ge> (\<beta>\<guillemotleft>\<lambda>a b c. let (args, y) = \<alpha> a b in add (foldr (\<lambda>x. plus (mulA (fst x) (snd x))) (zip args c) zero) y,s\<guillemotright>)"
      hence main: " polyEval \<beta> (foldr (\<lambda>ct. op + (fst ct * Pright \<alpha> (snd ct))) (zip vp ts) polyZero)
	       \<ge> (foldr (\<lambda>x. plus (mulA (fst x) (snd x))) (zip vp (map (interpretationOfTerm \<beta> (\<lambda>a b c. let (args, y) = \<alpha> a b 
                  in add (foldr (\<lambda>x. plus (mulA (fst x) (snd x))) (zip args c) zero) y)) ts)) zero)
               \<and> isPositive(polyC(foldr (\<lambda>ct. op + (fst ct * Pright \<alpha> (snd ct))) (zip vp ts) polyZero))" (is "polyEval \<beta> ?PRightVP \<ge> ?PInterVP \<and> _")
      proof (induct vp arbitrary:ts)
	case Nil
	thus ?case by (cases ts, auto simp:polyEvalVP.simps geqRefl posZeroD)
      next
	case (Cons n vpp)
	thus ?case proof (cases ts, simp add:polyEvalVP.simps geqRefl posZeroD)
	  case (Cons t tss)
	  from prems have ind1: "polyEval \<beta> (Pright \<alpha> t) \<ge> (\<beta>\<guillemotleft>\<lambda>a b c. let (args, y) = \<alpha> a b in add (foldr (\<lambda> x. plus (mulA (fst x) (snd x))) (zip args c) zero) y,t\<guillemotright>)" by simp
	  from prems have ind2: "polyEval \<beta>  (foldr (\<lambda>ct. op + (fst ct * Pright \<alpha> (snd ct))) (zip vpp tss) polyZero) \<ge> 
	    foldr (\<lambda>x. plus (mulA (fst x) (snd x))) (zip vpp (map (interpretationOfTerm \<beta> (\<lambda>a b c. let (args, a) = \<alpha> a b
            in add (foldr (\<lambda>x. plus (mulA (fst x) (snd x))) (zip args c) zero) a)) tss)) zero" by simp
	  from prems have ind3: "isPositive(polyC (foldr (\<lambda>ct. op + (fst ct * Pright \<alpha> (snd ct))) (zip vpp tss) polyZero))" by simp
	  from PrightPositive[where t = t and \<alpha> = \<alpha>] posMulD have fact1: "isPositive(polyC (n * Pright \<alpha> t))" by (cases "Pright \<alpha> t", auto)
	  let ?exp1 = "polyEval \<beta> (n * Pright \<alpha> t + (foldr (\<lambda>ct. op + (fst ct * Pright \<alpha> (snd ct))) (zip (vpp) (tss)) polyZero))"
	  let ?exp2 = "plus (polyEval \<beta> (n * Pright \<alpha> t)) (polyEval \<beta> (foldr (\<lambda>ct. op + (fst ct * Pright \<alpha> (snd ct))) (zip (vpp) (tss)) polyZero))"
	  let ?exp3 = "(foldr (\<lambda>x. plus (mulA (fst x) (snd x))) (zip (n # vpp)
            (map (interpretationOfTerm \<beta> (\<lambda>a b c. let (args, y) = \<alpha> a b in add (foldr (\<lambda>x. plus (mulA (fst x) (snd x)))
            (zip args c) zero) y)) (t # tss))) zero)"

	  have "?exp1 = ?exp2" using polySumSound2[where ?p = "n * Pright \<alpha> t"] fact1 ind3 by auto
	  have "?exp2 \<ge> ?exp3" by (simp, rule plusMono, rule mulAMono, simp add: ind1, simp add: ind2)
	  with `?exp1 = ?exp2` have "?exp1 \<ge> ?exp3" by simp
	  with Cons fact1 sumPositive ind3 show ?thesis by auto 
	qed
      qed
      thus "polyEval \<beta> (case ?PRightVP + Poly [] c of Poly vp d \<Rightarrow> Poly vp (maxZero d)) 
            \<ge> add ?PInterVP c "
      proof (cases "?PRightVP")
	case (Poly vpp d)
	from main have "polyEval \<beta> ?PRightVP \<ge> ?PInterVP" and "isPositive(polyC ?PRightVP)" by auto
	with Poly obtain PInterVP where eq: "?PInterVP = PInterVP" and geq: "polyEval \<beta> (Poly vpp d) \<ge> PInterVP" and pos: "isPositive(d)" by auto
	from geq pos have part1: "polyEval \<beta> (Poly vpp d + Poly [] c) \<ge> add PInterVP c"	  
	  by (simp add: posAddAddD addMono)	
	have "polyEval \<beta> (case Poly vpp d + Poly [] c of Poly vp d \<Rightarrow> Poly vp (maxZero d)) \<ge> add PInterVP c"
	proof (rule geqTrans[where a' = "add (polyEvalVP \<beta> vpp) (addD d c)"], simp add: maxZeroAdd)
	  show "add (polyEvalVP \<beta> vpp) (addD d c) \<ge> add PInterVP c" using part1 by simp
	qed
	with Poly eq show ?thesis by simp
      qed
    qed  
  qed
qed



declare Pright.simps Pleft.simps[simp del]


definition polyGe :: "('v,'b,'d)poly \<Rightarrow> ('v,'b,'d)poly \<Rightarrow> bool" (infix "\<ge>" 10)
  where "(p1 \<ge> p2) = (\<forall> \<beta>. polyEval \<beta> p1 \<ge> polyEval \<beta> p2)"
definition polyGt :: "('v,'b,'d)poly \<Rightarrow> ('v,'b,'d)poly \<Rightarrow> bool" (infix ">" 10)
  where "(p1 > p2) = (\<forall> \<beta>. polyEval \<beta> p1 > polyEval \<beta> p2)"


lemma polyApproxSoundGe : assumes orient: "Pleft \<alpha> s \<ge> Pright \<alpha> t"
  shows "(s,t) \<in> interNS (pol \<alpha>)"
proof (unfold interNS_def, simp, rule allI)
  fix \<beta> 
  from orient have "polyEval \<beta> (Pleft \<alpha> s) \<ge> polyEval \<beta> (Pright \<alpha> t)" unfolding polyGe_def by auto
  with PleftSound[where t = s and \<beta> = \<beta> and \<alpha> = \<alpha>] PrightSound[where t = t and \<beta> = \<beta> and \<alpha> = \<alpha>]
  show "(\<beta>\<guillemotleft>pol \<alpha>,s\<guillemotright>) \<ge> (\<beta>\<guillemotleft>pol \<alpha>,t\<guillemotright>)" using geqTrans by blast 
qed

lemma polyApproxSoundGt : assumes orient: "Pleft \<alpha> s > Pright \<alpha> t"
  shows "(s,t) \<in> interS (pol \<alpha>)"
proof (unfold interS_def, simp, rule allI)
  fix \<beta> 
  from orient have "polyEval \<beta> (Pleft \<alpha> s) > polyEval \<beta> (Pright \<alpha> t)" unfolding polyGt_def by auto
  with PleftSound[where t = s and \<beta> = \<beta> and \<alpha> = \<alpha>] PrightSound[where t = t and \<beta> = \<beta> and \<alpha> = \<alpha>]
  show "(\<beta>\<guillemotleft>pol \<alpha>,s\<guillemotright>) > (\<beta>\<guillemotleft>pol \<alpha>,t\<guillemotright>)" using geqGt by blast
qed



(* now we come to checking that two terms are in relation  *)

fun checkVPLists :: "('b \<times> 'v) list \<Rightarrow> ('b \<times> 'v) list \<Rightarrow> 'v monad"
  where "checkVPLists _ [] = OK"
     |  "checkVPLists vp ((n,x) # vq) = (if n = zeroB then checkVPLists vp vq else (case lookupVar x vp of None \<Rightarrow> error x | Some (m, vpp) \<Rightarrow> checkB (geqB m n) x \<guillemotright> checkVPLists vpp vq))"

lemma checkVPListsSound[simp] : "isOK(checkVPLists vp vq) \<Longrightarrow> polyEvalVP \<beta> vp \<ge> polyEvalVP \<beta> vq"
proof (induct vq arbitrary: vp)
  case Nil thus ?case by (simp add: polyEvalVP.simps zeroLeast)
next
  case (Cons nx vqq)
  show ?case 
  proof (cases nx)
    case (Pair n x)
    with Cons show ?thesis 
    proof (cases "n = zeroB", simp add: polyEvalVP.simps)
      case False
      with Cons Pair show ?thesis proof (cases "lookupVar x vp", simp add: polyEvalVP.simps)
	case (Some mvpp)
	show ?thesis 
	proof (cases mvpp)
	  case (Pair m vpp)
	  show ?thesis 
	  proof (cases "geqB m n")
	    case False with Cons Pair Some `n \<noteq> zeroB` `nx = (n,x)` show ?thesis by simp
	  next
	    case True with Cons Pair Some `n \<noteq> zeroB` `nx = (n,x)` have "isOK(checkVPLists vpp vqq)" by simp
	    with Cons have ind: "polyEvalVP \<beta> vpp \<ge> polyEvalVP \<beta> vqq" by simp
	    from Some Pair have lookup: "lookupVar x vp = Some (m,vpp)" by simp
	    let ?exp0 = "plus (mulA m (\<beta> x)) (polyEvalVP \<beta> vpp)"
	    let ?exp1 = "plus (mulA n (\<beta> x)) (polyEvalVP \<beta> vqq)"
	    let ?exp2 = "polyEvalVP \<beta> (nx # vqq)" 
	    have "?exp1 = ?exp2" using `nx = (n,x)` by (simp add: polyEvalVP.simps)
	    have "?exp0 \<ge> ?exp1" by (rule plusMono, rule mulAMonoLeft, rule `geqB m n`, rule ind)
	    from lookup have "?exp0 = polyEvalVP \<beta> vp"
	    proof (induct vp arbitrary: m vpp)
	      case Nil thus ?case by simp
	    next
	      case (Cons cy vp)
	      show ?case
	      proof (cases cy)
		case (Pair c y)
		with Cons show ?thesis proof (cases "x = y", simp add: polyEvalVP.simps)
		  case False
		  with Cons Pair show ?thesis proof (cases "lookupVar x vp", simp)
		    case (Some dvppp)
		    show ?thesis proof (cases dvppp)
		      case (Pair d vppp)
		      have "plus (mulA m (\<beta> x)) (plus (mulA c (\<beta> y)) (foldr (\<lambda>bv. plus (mulA (fst bv) (\<beta> (snd bv)))) vppp zero)) = 
			    plus (mulA c (\<beta> y)) (plus (mulA m (\<beta> x)) (foldr (\<lambda>bv. plus (mulA (fst bv) (\<beta> (snd bv)))) vppp zero))"
			using plusAssoc plusComm by auto		      
		      with Pair Cons False `cy = (c,y)` Some show ?thesis by (auto simp: polyEvalVP.simps)
		    qed
		  qed
		qed
	      qed
	    qed
	    with `?exp1 = ?exp2` `?exp0 \<ge> ?exp1`  show ?thesis by simp
	  qed
	qed
      qed
    qed
  qed
qed



fun checkPolyGe :: "('v :: sshow,'b,'d) poly \<Rightarrow> ('v,'b,'d)poly \<Rightarrow> sstring list monad"
  where "checkPolyGe (Poly vp c) (Poly vq d) = (
   (checkB (geqD c d)
      [''the constant of the left polynomial is too small'']
  \<guillemotright> checkVPLists vp vq 
     \<leftarrow> (\<lambda> x. [ ''variable '' +++ sshow x +++  '' is problematic'']))
  \<leftarrow> (\<lambda> s. (sConcat [''Could not orient '', sshow (Poly vp c), '' >= '', sshow (Poly vq d)]) # s))"



lemma checkPolyGeSound[simp] : "isOK(checkPolyGe p q) \<Longrightarrow> p \<ge> q"
proof -
  assume check: "isOK(checkPolyGe p q)"
  show "p \<ge> q" unfolding polyGe_def
  proof
    fix \<beta> 
    show "polyEval \<beta> p \<ge> polyEval \<beta> q"
    proof (cases p)
      case (Poly vp c)
      show ?thesis
      proof (cases q)
	case (Poly vq d)	
	from prems have one: "polyEvalVP \<beta> vp \<ge> polyEvalVP \<beta> vq" and two: "geqD c d" by auto
	have "polyEval \<beta> (Poly vp c) \<ge> polyEval \<beta> (Poly vq d)" 
	proof (rule geqTrans)
	  show "polyEval \<beta> (Poly vp c) \<ge> polyEval \<beta> (Poly vp d)"
	    using two geqDMono by simp
	next
	  show "polyEval \<beta> (Poly vp d) \<ge> polyEval \<beta> (Poly vq d)"
	    using one addMono by simp
	qed
	with prems show ?thesis by simp
      qed
    qed
  qed
qed

declare checkPolyGe.simps[simp del]

fun checkPolyGt :: "('v :: sshow,'b,'d) poly \<Rightarrow> ('v,'b,'d)poly \<Rightarrow> sstring list monad"
  where "checkPolyGt (Poly vp c) (Poly vq d) = (
   (checkB (gtD c d)
      [''the constant of the left polynomial is too small'']
  \<guillemotright> checkVPLists vp vq 
     \<leftarrow> (\<lambda> x. [ ''variable '' +++ sshow x +++  '' is problematic'']))
  \<leftarrow> (\<lambda> s. (sConcat [''Could not orient '', sshow (Poly vp c), '' > '', sshow (Poly vq d)]) # s))"



lemma checkPolyGtSound[simp] : "isOK(checkPolyGt p q) \<Longrightarrow> p > q"
proof -
  assume check: "isOK(checkPolyGt p q)"
  show "p > q" unfolding polyGt_def
  proof
    fix \<beta> 
    show "polyEval \<beta> p > polyEval \<beta> q"
    proof (cases p)
      case (Poly vp c)
      show ?thesis
      proof (cases q)
	case (Poly vq d)	
	from prems have one: "polyEvalVP \<beta> vp \<ge> polyEvalVP \<beta> vq" and two: "gtD c d" by auto
	have "polyEval \<beta> (Poly vp c) > polyEval \<beta> (Poly vq d)" 	  
	proof (simp, rule geqGt, rule geqRefl)
	  show "add (polyEvalVP \<beta> vp) c > add (polyEvalVP \<beta> vp) d"
	    using two  gtDMono by simp
	next
	  show "add (polyEvalVP \<beta> vp) d \<ge> add (polyEvalVP \<beta> vq) d"
	    using one addMono by simp
	qed
	with prems show ?thesis by simp
      qed
    qed
  qed
qed

declare checkPolyGt.simps[simp del]

fun genIndices :: "nat \<Rightarrow> int \<Rightarrow> ('f,sstring) term  list"
  where "genIndices 0 _ = []" 
  |  "genIndices (Suc n) i = (Var (''x_'' +++ (sshow i)) # genIndices n (i+1))"
fun genTerm :: "'f \<Rightarrow> nat \<Rightarrow> ('f,string)term" where "genTerm f n = Fun f (genIndices n 0)"

fun "showInterpretation" :: "('c,'b,'d)polyinter \<Rightarrow> 'c \<Rightarrow> nat \<Rightarrow> sstring"
  where "showInterpretation \<alpha> f n = (sshow (Pleft \<alpha> (genTerm f n)))"



definition checkNSPolo :: "('c,'b,'d)polyinter \<Rightarrow> ('c :: sshow,'v :: sshow)rule \<Rightarrow> sstring list monad"
  where "checkNSPolo \<alpha> lr \<equiv> checkPolyGe (Pleft \<alpha>  (fst lr)) (Pright \<alpha> (snd lr)) 
      \<leftarrow> (\<lambda> s. (sConcat [ ''Could not orient '' , sshow (fst lr), '' >= '', sshow (snd lr)]) # s)
"

lemma checkNSPoloSound[simp] : "isOK(checkNSPolo \<alpha> lr) \<Longrightarrow> lr \<in> interNS (pol \<alpha>)"
  unfolding checkNSPolo_def by (cases lr, auto simp: polyApproxSoundGe)


definition checkSPolo :: "('c,'b,'d)polyinter \<Rightarrow> ('c :: sshow,'v :: sshow)rule \<Rightarrow> sstring list monad"
  where "checkSPolo \<alpha> lr \<equiv> checkPolyGt (Pleft \<alpha>  (fst lr)) (Pright \<alpha> (snd lr)) 
      \<leftarrow> (\<lambda> s. (sConcat [ ''Could not orient '' , sshow (fst lr), '' > '', sshow (snd lr)]) # s)
"

lemma checkSPoloSound[simp] : "isOK(checkSPolo \<alpha> lr) \<Longrightarrow> lr \<in> interS (pol \<alpha>)"
  unfolding checkSPolo_def by (cases lr, auto simp: polyApproxSoundGt)



fun createRepresentation :: "('c :: sshow ,'b,'d)polyinter \<Rightarrow> ('c \<times> nat) list \<Rightarrow> sstring list"
  where "createRepresentation \<alpha> fs =  (''polynomial interpretation: '' #
             map (\<lambda> (g,n) . sConcat [''Pol('', sshow g, ''/'', sshow n, '') = '', showInterpretation \<alpha> g n]) (remdups fs))"
end 



(* and hence, they also form a reduction pair *)

interpretation natPolyRedpair:  redpair [ "wfcarrier.interS (op > :: nat \<Rightarrow> nat \<Rightarrow> bool) (polyinter.pol (op +) 0 add (op * ) \<alpha>)" "wfcarrier.interNS (op \<ge> :: nat \<Rightarrow> nat \<Rightarrow> bool) (polyinter.pol (op +) 0 add (op * ) \<alpha> )" ]
by (unfold_locales, auto)


instantiation cint :: sshow
begin
fun sshow_cint :: "cint \<Rightarrow> sstring" 
where "sshow_cint (Neg 0) = ''0''"
    | "sshow_cint (Pos 0) = ''0''"
    | "sshow_cint (Neg (Suc n)) = ''-'' @ (sshow (Suc n))"
    | "sshow_cint (Pos (Suc n)) = sshow (Suc n)"
instance ..
end


fun addI :: "cint \<Rightarrow> cint \<Rightarrow> cint" 
where "addI (Neg x) (Neg y)  = Neg (x + y)"
  |   "addI (Pos x) (Pos y)  = Pos (x + y)"
  |   "addI (Neg x) (Pos y)  = (if x \<le> y then Pos (y - x) else Neg (x - y))"
  |   "addI (Pos x) (Neg y)  = (if y \<le> x then Pos (x - y) else Neg (y - x))"

fun mul :: "nat \<Rightarrow> cint \<Rightarrow> cint"
where "mul n (Neg m) = Neg (n*m)"
   |  "mul n (Pos m) = Pos (n*m)"

fun geqCint :: "cint \<Rightarrow> cint \<Rightarrow> bool" 
where "geqCint (Pos x) (Pos y) = (x \<ge> y)"
   |  "geqCint (Pos x) (Neg y) = True"
   |  "geqCint (Neg x) (Neg y) = (x \<le> y)"
   |  "geqCint (Neg x) (Pos y) = (x = 0 \<and> y = 0)"

fun gtCint :: "cint \<Rightarrow> cint \<Rightarrow> bool" 
where "gtCint (Pos x) (Pos y) = (x > y)"
   |  "gtCint (Pos x) (Neg y) = (x > 0)" (* this is due to the fact that we have to consider the positive part mainly *)
   |  "gtCint (Neg x) (Neg y) = False" (* this is due to the fact that we have to consider the positive part mainly *)
   |  "gtCint (Neg x) (Pos y) = False"
 
lemma geqCintMono: "geqCint d d' \<Longrightarrow> add a d \<ge> add a d'"
  by  (cases "d", cases "d'", simp, simp, cases "d'", auto)

lemma gtCintMono: "gtCint d d' \<Longrightarrow> add a d > add a d'"
  by  (cases "d", cases "d'", simp, simp, cases "d'", auto)

lemma[simp]: "n * add m c = add (n * m) (mul n c)"
  by (cases c, auto simp: nat_distrib)

fun cintMaxZero :: "cint \<Rightarrow> cint"
where "cintMaxZero (Pos x) = (Pos x)" | "cintMaxZero (Neg x) = Pos 0"

lemma cintMaxZeroAddZero: "add 0 d \<ge> add 0 (cintMaxZero d)"
  by (cases d, auto)
lemma cintMaxZeroAdd: "add n (cintMaxZero d) \<ge> add n d"
  by (cases d, auto)


lemma addPlusAddI : " (add a d) +  (add a' d') \<ge> add (a + a') (addI d d')"
  by (cases d,cases d',simp,simp, cases d', auto)
lemma addAddI : "add (add a d) d' \<ge> add a (addI d d')"
  by (cases d,cases d',simp,simp, cases d', simp, simp)
lemma posAddAddI : "cintMaxZero d = d \<Longrightarrow> add a (addI d d') = add (add a d) d'"
  by (cases d,cases d',simp,simp, cases d', simp, simp)
lemma posMulD: "cintMaxZero d = d \<Longrightarrow> cintMaxZero (mul b d) = mul b d"
  by (cases d, auto)
lemma posAddI: "\<lbrakk>cintMaxZero d = d; cintMaxZero d' = d'\<rbrakk> \<Longrightarrow> cintMaxZero (addI d d') = addI d d'"
  by (cases d, simp, cases d', auto)
lemma posCintMaxZero: "cintMaxZero (cintMaxZero x) = cintMaxZero x"
  by (cases x, auto)
lemma posPlusAdd: "\<lbrakk>cintMaxZero d = d; cintMaxZero d' = d'\<rbrakk>
       \<Longrightarrow> add (a + a') (addI d d') = add a d + add a' d'"
  by (cases d, simp, cases d', simp, simp)

interpretation natPolyInterImpl : polyinterImpl [ "op > :: nat \<Rightarrow> nat \<Rightarrow> bool" "op \<ge> :: nat \<Rightarrow> nat \<Rightarrow> bool" "op + :: nat \<Rightarrow> nat \<Rightarrow> nat" "0 :: nat" "add :: nat \<Rightarrow> cint \<Rightarrow> nat" "op * :: nat \<Rightarrow> nat \<Rightarrow> nat" "addI" "Pos 0" "mul" "op * :: nat \<Rightarrow> nat \<Rightarrow> nat" "op + :: nat \<Rightarrow> nat \<Rightarrow> nat" "1" "cintMaxZero" "0 :: nat" "op \<ge> :: nat \<Rightarrow> nat \<Rightarrow> bool" geqCint gtCint] 
  by (unfold_locales, auto simp: nat_distrib cintMaxZeroAddZero addPlusAddI addAddI posMulD posAddI posCintMaxZero posPlusAdd geqCintMono gtCintMono, auto simp: posAddAddI cintMaxZeroAdd)

(* it remains to create some polynomial interpretation from a finite list *)

types 'f natCintPolyinter =  "'f \<Rightarrow> nat \<Rightarrow> (nat list) \<times> cint"



fun createPolyInter :: "('f \<times> (cint \<times> nat list)) list \<Rightarrow> 'f natCintPolyinter"
  where "createPolyInter [] f n = ([], Pos 0)"
     |  "createPolyInter ((g,(i,coeffs)) # rest) f n = (if (f = g) \<and> (n = length coeffs) then (coeffs, i) else createPolyInter rest f n)"

datatype 'f RedPair = Polo "('f \<times> (nat \<times> nat list)) list" | NegPolo "('f \<times> (cint \<times> nat list)) list"

fun redPairToPrePolyInter :: "'f RedPair \<Rightarrow> ('f \<times> (cint \<times> nat list)) list"
  where "redPairToPrePolyInter (Polo inter) = map (\<lambda> (f, (i,coeffs)). (f, (Pos i, coeffs))) inter" 
      | "redPairToPrePolyInter (NegPolo inter) = inter"


(* this routine should be inside the locale! *)
(* moreover, one might introduce redpairImpl again, if code-generation from locales works fine *)
fun createRedPairProc :: "'f RedPair \<Rightarrow> ('f :: sshow ,'v :: sshow)trsL \<Rightarrow> ('f, 'v)trsL \<Rightarrow> ('f, 'v)trsL \<Rightarrow> 
        sstring list monad"
where "createRedPairProc interpr P rP R = (let \<alpha> = createPolyInter (redPairToPrePolyInter interpr) in 
  checkRedPairProc 
           (natPolyInterImpl.checkSPolo \<alpha>)
           (natPolyInterImpl.checkNSPolo \<alpha>)
           (''polynomial order'') 
           (natPolyInterImpl.createRepresentation \<alpha> (remdups (map (\<lambda> (g,il) . (g, length (snd il))) (redPairToPrePolyInter interpr))))
           P rP R)" 


lemma createRedPairProcSound[simp] : fixes P rP R
assumes check: "isOK(createRedPairProc redp P rP R)"
   shows "finite_dpp (set rP, set R) \<Longrightarrow> finite_dpp (set P, set R)"
using check
proof (simp add: Let_def)
  assume fin: "finite_dpp (set rP, set R)" 
  let ?alpha = "(createPolyInter (redPairToPrePolyInter redp))"
  let ?checkNS = "polyinterImpl.checkNSPolo addI (Pos 0) mul
            op * op + (Suc 0) cintMaxZero 0 (\<lambda>x y. y \<le> x) geqCint ?alpha"
  let ?checkS = "polyinterImpl.checkSPolo addI (Pos 0) mul
            op * op + (Suc 0) cintMaxZero 0 (\<lambda>x y. y \<le> x) gtCint ?alpha"
  let ?succsim = "wfcarrier.interNS (\<lambda> x y. y \<le> x) (polyinter.pol op + 0 add op * ?alpha)"
  let ?succ = "wfcarrier.interS (\<lambda> x y. y < x) (polyinter.pol op + 0 add op * ?alpha)"
  show ?thesis 
  proof (cases "computeSplit (\<lambda> dp. isOK(?checkS dp)) P")
    case (Pair Sp NSp)
    let ?NSp = "set NSp \<inter> set rP"
    from Pair check have cond:
      "set P - set NSp \<subseteq> ?succ \<and> 
      ?NSp \<inter> set P \<subseteq> ?succsim \<union> ?succ   \<and> 
      set R \<subseteq> ?succsim     \<and>
      set NSp \<subseteq> set rP"
      by (auto simp: Let_def mem_iff)
    hence id: "set NSp = ?NSp" and stri: "set P - ?NSp \<subseteq> ?succ" by (auto simp: mem_iff)
    from fin have fin2: "finite_dpp (?NSp, set R)" unfolding finite_dpp_def by (auto simp:mem_iff)
    show "finite_dpp (set P, set R)" 
    proof (rule natPolyRedpair.redPairProcessorDpp)
      show "finite_dpp (?NSp, set R)" using fin2 .
    next
      from stri show "set P - ?NSp \<subseteq> ?succ" .
    next
      from cond id show "?NSp \<inter> set P \<subseteq> ?succsim \<union> ?succ" by simp
    next
      from cond show "set R \<subseteq> ?succsim" by simp
    qed
  qed
qed
    
declare createRedPairProc.simps[simp del]


*)

end




