(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Show imports Util (* Code_Char_chr *) begin

types sstring = string

class "show" =
  fixes "show" :: "'a \<Rightarrow> string"

definition unquote :: "sstring \<Rightarrow> sstring"
  where "unquote x \<equiv> filter (\<lambda> c. \<not> c = CHR ''\"'') x"
definition showF :: "'a :: show \<Rightarrow> sstring"
  where "showF x \<equiv> unquote (show x)"

class sshow =
  fixes sshow :: "'a \<Rightarrow> sstring"

abbreviation sAppend :: "sstring \<Rightarrow> sstring \<Rightarrow> sstring" (infixl "+++" 10)
  where "sAppend s t \<equiv> (s @ t)"


abbreviation sConcat :: "sstring list \<Rightarrow> sstring" where "sConcat \<equiv> concat"


definition sEmpty :: "sstring" where "sEmpty = []"


instantiation bool  :: sshow
begin
fun sshow_bool where "sshow_bool True = ''True''" | "sshow_bool False = ''False''"
instance by intro_classes
end

instantiation char  :: sshow
begin
fun sshow_char where "sshow_char x = ([x] :: string)"
instance by intro_classes
end

instantiation list :: ("sshow") "sshow"
begin
fun sshow_list where "sshow xs = sConcat (map sshow xs)"
instance by intro_classes
end

definition newlineC :: char where "newlineC \<equiv> Char Nibble0 NibbleA"
definition tabC :: char where "tabC \<equiv> Char Nibble0 Nibble9"
definition tab :: string where "tab \<equiv> [tabC]"
definition newline :: string where "newline \<equiv> [newlineC]"
 

instantiation * :: (sshow,sshow) sshow
begin
fun sshow_prod where "sshow_prod (x,y) = (concat [sshow x, '' -> '', sshow y])"
instance by intro_classes
end




fun showList  :: "sstring list \<Rightarrow> string"
 where "showList xs = concat (map (\<lambda> x. x @ newline) xs)"


instantiation nat :: sshow
begin
fun digit2string :: "nat \<Rightarrow> string"
where "digit2string n = 
   (if n = 0 then ''0'' else
   (if n = 1 then ''1'' else
   (if n = 2 then ''2'' else
   (if n = 3 then ''3'' else
   (if n = 4 then ''4'' else
   (if n = 5 then ''5'' else
   (if n = 6 then ''6'' else
   (if n = 7 then ''7'' else
   (if n = 8 then ''8'' else
   (if n = 9 then ''9'' else ''not a digit''))))))))))"
  
fun sshow_nat :: "nat \<Rightarrow> string" 
where "sshow_nat n = (if n < 10 then digit2string n else sshow_nat (n div 10) @ digit2string (n mod 10))"
instance by intro_classes
end

instantiation cint :: sshow
begin
fun sshow_cint where
  "sshow_cint (Pos n) = sshow n"
| "sshow_cint (Neg n) = (if n = 0 then sshow n else (''-'' @ sshow n))"
instance ..
end

instantiation int :: sshow
begin
fun sshow_int where "sshow_int i = sshow (intToCint i)"
instance .. 
end


text {* a fresh name generator for creation
 of rules f(x_1,..,x_n) \<rightarrow> ...
 or for renaming variables apart, 
 etc.
*}

definition fresh_name_gen where "fresh_name_gen f \<equiv> \<forall> n xs. length (f n xs) = n \<and> set (f n xs) \<inter> set xs = {} \<and> distinct (f n xs)"

lemma digit2string_inj : assumes "x < y \<and> y < 10"
  shows "digit2string x \<noteq> digit2string y"
using assms
by ( 
  cases "y=1", simp,
  cases "y=2", simp,
  cases "y=3", simp,
  cases "y=4", simp,
  cases "y=5", simp,
  cases "y=6", simp,
  cases "y=7", simp,
  cases "y=8", simp, 
  cases "y=9", simp, 
  arith
)

lemma digit2string_inj_2 : assumes "x < 10"
  shows "length (digit2string x) = 1"
using assms
by ( 
  cases "x=0", simp,
  cases "x=1", simp,
  cases "x=2", simp,
  cases "x=3", simp,
  cases "x=4", simp,
  cases "x=5", simp,
  cases "x=6", simp,
  cases "x=7", simp,
  cases "x=8", simp,
  cases "x=9", simp, arith)

declare digit2string.simps[simp del]

lemma digit2string_inj_1 : assumes "x < 10" and "y < 10" and "x \<noteq> y"
  shows "digit2string x \<noteq> digit2string y"
proof
  assume ass: "digit2string x = digit2string y"
  with assms have  "\<exists> xx yy :: nat. xx < 10 \<and> yy < 10 \<and> xx < yy \<and> digit2string xx = digit2string yy" 
  proof (cases "x < y", blast)
    case False
    with `x \<noteq> y` have "y < x" by arith
    with assms ass have "y < 10 \<and> x < 10 \<and> y < x \<and> digit2string y = digit2string x" by auto
    thus ?thesis by blast
  qed
  from this obtain xx yy :: nat where cond: "xx < 10 \<and> yy < 10 \<and> xx < yy \<and> digit2string xx = digit2string yy" by blast
  thus False using digit2string_inj by auto
qed


lemma show_nat_len_1 : fixes x :: nat 
  shows "length (sshow x) > 0"
proof (cases "x < 10", simp add: digit2string_inj_2)
  assume ass: "\<not> x < 10"
  hence "sshow x = sshow (x div 10) @ digit2string (x mod 10)" by simp
  hence "length (sshow x) = length (sshow (x div 10)) + length (digit2string (x mod 10))" by (simp del: sshow_nat.simps)
  with ass show ?thesis by (simp del: sshow_nat.simps add: digit2string_inj_2)
qed

lemma show_nat_len_2 : fixes x :: nat 
  assumes one: "length (sshow x) = 1" 
  shows "x < 10"
proof (rule ccontr)
  assume "\<not> x < 10"
  hence "sshow x = sshow (x div 10) @ digit2string (x mod 10)" by simp
  hence "length (sshow x) = length (sshow (x div 10)) + length (digit2string (x mod 10))" by (simp del: sshow_nat.simps)
  with one have "length (sshow (x div 10)) = 0" by (simp del: sshow_nat.simps add: digit2string_inj_2)
  with show_nat_len_1 show False by best
qed

lemma show_nat_len_3 : fixes x :: nat 
  assumes one: "length (sshow x) > 1"
  shows "\<not> x < 10"
proof 
  assume ass: "x < 10"
  hence "sshow x = digit2string x" by simp
  with digit2string_inj_2 ass have "length (sshow x) = 1" by (auto simp del: sshow_nat.simps)
  with one show False by arith
qed



lemma show_nat_inj : "inj (sshow :: nat \<Rightarrow> string)"
unfolding inj_on_def
proof (clarify)
  fix x y :: nat

  assume "sshow x = sshow y"
  from this obtain n where "length (sshow x) = n \<and> sshow x = sshow y" by blast
  thus "x = y" 
  proof (induct n arbitrary: x y)
    case 0
    fix x y :: nat
    assume ass: "length (sshow x) = 0 \<and> sshow x = sshow y"
    thus ?case using show_nat_len_1 by best
  next
    case (Suc n)
    hence eq: "sshow x = sshow y" by blast
    from Suc have len: "length (sshow x) = Suc n" by blast
    show ?case
    proof (cases n)
      case 0
      with Suc have one: "length (sshow x) = 1 \<and> length (sshow y) = 1" by (auto simp del: sshow_nat.simps)
      with show_nat_len_2 have ten: "x < 10 \<and> y < 10" by (auto simp del: sshow_nat.simps)      
      show ?thesis 
      proof (cases "x = y", simp)
	case False
	with ten digit2string_inj_1 have "digit2string x \<noteq> digit2string y" by auto
	with ten eq show ?thesis by auto
      qed
    next
      case (Suc m)
      with len eq have "length (sshow x) = Suc (Suc m) \<and> length (sshow y) = Suc (Suc m)" by (auto simp del: sshow_nat.simps)
      hence "length (sshow x) > 1 \<and> length (sshow y) > 1" by (auto simp del: sshow_nat.simps)
      hence large: "\<not> x < 10 \<and> \<not> y < 10" using show_nat_len_3 by (auto simp del: sshow_nat.simps)
      let ?e = "\<lambda> x :: nat. sshow (x div 10) @ digit2string (x mod 10)"
      from large have id: "sshow x = ?e x \<and> sshow y = ?e y" by auto
      from digit2string_inj_2 have one: "length (digit2string (x mod 10)) = 1 \<and> length (digit2string (y mod 10)) = 1" by auto
      from one have "\<exists> xx. digit2string (x mod 10) = [xx]" by (cases "digit2string (x mod 10)", simp, cases "tl (digit2string (x mod 10))", simp, simp)
      from this obtain xx where xx: "digit2string (x mod 10) = [xx]" ..
      from one have "\<exists> xx. digit2string (y mod 10) = [xx]" by (cases "digit2string (y mod 10)", simp, cases "tl (digit2string (y mod 10))", simp, simp)
      from this obtain yy where yy: "digit2string (y mod 10) = [yy]" ..

      from one len id have shorter: "length (sshow (x div 10)) = n" by (auto simp del: sshow_nat.simps)
      from eq id xx yy
      have "(sshow (x div 10) +++ [xx]) = (sshow (y div 10) +++ [yy])"  by (auto simp del: sshow_nat.simps)
      hence "(sshow (x div 10) = sshow (y div 10)) \<and> xx = yy" by simp
      hence nearly: "sshow (x div 10) = sshow (y div 10) \<and> digit2string (x mod 10) = digit2string (y mod 10)" using xx yy by simp
      with shorter `length (sshow (x div 10)) = n \<and> sshow (x div 10) = sshow (y div 10) \<Longrightarrow> x div 10 = y div 10`
      have div: "x div 10 = y div 10" by blast
      from nearly digit2string_inj_1 have mod:  "x mod 10 = y mod 10"
	by (cases "x mod 10 = y mod 10", simp, simp)
      have "x = 10 * (x div 10) + (x mod 10)" by auto
      also have "\<dots> = 10 * (y div 10) + (y mod 10)" using div mod by auto
      also have "\<dots> = y" by auto
      finally show ?thesis .
    qed
  qed
qed

declare sshow_nat.simps[simp del]

definition fresh_strings :: "string \<Rightarrow> nat \<Rightarrow> nat \<Rightarrow> string list \<Rightarrow> string list" 
where "fresh_strings name offset n used \<equiv> take n (filter (\<lambda> s. \<not> s mem used) (map (\<lambda> i. name @ sshow (i + offset)) [0..< n + length used]))"

lemma distinct_mem: assumes "distinct xs"
  shows "length xs - length ys \<le> length (filter (\<lambda> s. \<not> s mem ys) xs)"
using assms 
proof (induct xs arbitrary: ys, simp)
  case (Cons x xs)
  hence dist: "distinct xs" by auto
  show ?case
  proof (cases "x mem ys")
    case False
    from dist Cons have "length xs - length ys \<le> length [s\<leftarrow>xs . \<not> s mem ys]" by auto
    with False show ?thesis by auto
  next
    case True
    let ?ys = "remove1 x ys"
    from True Cons have nomem: "x \<notin> set xs" by auto
    from dist Cons have len1: "length xs - length ?ys \<le> length [s\<leftarrow>xs . \<not> s mem ?ys]" by auto
    from True have len2: "length ys - 1 = length ?ys" by (simp add: mem_iff length_remove1)
    from True obtain n where "length ys = Suc n" by (cases ys, auto) 
    with len2 have len3: "length ys = Suc (length ?ys)" by auto
    with len1 have len: "length (x # xs) - length ys \<le> length [s\<leftarrow>xs. \<not> s mem ?ys]" by auto
    from nomem have "[s\<leftarrow>xs . \<not> s mem remove1 x ys] = [s\<leftarrow>xs . \<not> s mem ys]" 
    proof (induct xs, simp)
      case (Cons xx xs)
      thus ?case by (cases "x = xx", simp, auto simp: mem_iff)
    qed
    with len show ?thesis by auto
  qed
qed
    

lemma "fresh_name_gen (fresh_strings name offset)"   
unfolding fresh_name_gen_def fresh_strings_def
proof (intro allI)
  fix n :: nat and used :: "string list"
  let ?is = "[0 ..< n + length used]"
  let ?fun = "\<lambda> i :: nat. name @ sshow (i + offset)"
  let ?filt = "\<lambda> s. \<not> s mem used"
  let ?map = "map ?fun ?is"
  let ?fil = "filter ?filt ?map"
  have two: "set (take n ?fil) \<inter> set used = {}" (is "?f \<inter> ?u = {}")
  proof 
    show "{} \<subseteq> ?f \<inter> ?u" by auto
  next
    show "?f \<inter> ?u \<subseteq> {}"
    proof
      fix x
      assume "x \<in> ?f \<inter> ?u"
      hence "x \<in> set ?fil" using set_take_subset[where n = n and xs = ?fil] by blast
      hence "?filt x" by auto
      with `x \<in> ?f \<inter> ?u` have False by (auto simp: mem_iff)
      thus "x \<in> {}" ..
    qed
  qed
  have mapDist: "distinct ?map" 
  proof (simp add: distinct_map inj_on_def, (intro allI, intro impI)+)
    fix x y :: nat
    assume "sshow (x + offset) = sshow (y + offset)"
    with show_nat_inj have "x + offset = y + offset" unfolding inj_on_def by force
    thus "x = y" by auto
  qed
  hence filDist: "distinct ?fil" by auto
  hence filFresh: "distinct (take n ?fil)" by auto
  have "length ?fil \<ge> n" using mapDist distinct_mem[where xs = ?map and ys = used] by auto
  hence "length (take n ?fil) = n" by auto

  with filFresh two show "length (take n ?fil) = n \<and> set (take n ?fil) \<inter> set used = {} \<and> distinct (take n ?fil)" by blast
qed


end


