(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory SubtermCriterion imports DpFramework begin

text {* Simple Projection. *}
types 'f sp = "'f \<Rightarrow> nat"

text {* A simple projection collapses a term to an argument depending on its
root symbol. *}
fun lift2term :: "'f sp \<Rightarrow> ('f,'v)term \<Rightarrow> ('f,'v)term"
where "lift2term p (Var x) = Var x"
    | "lift2term p (Fun f ss) = ss!(p f)"

lemma ctxt_closed_imp_commute_supt: assumes "ctxt_closed R" shows "R O supt \<subseteq> supt O R"
proof (rule subsetI2)
 fix s t assume "(s,t) \<in> R O supt"
 then obtain u where "(s,u) \<in> supt" and "(u,t) \<in> R" by auto
 from `(s,u) \<in> supt`[to_pred] have "s \<rhd> u" .
 obtain v where "(s,v) \<in> R" and "v \<rhd> t" using ctxt_closed_R_imp_suptp_R_distr[OF assms `s \<rhd> u` `(u,t) \<in> R`] by auto
 from `v \<rhd> t`[to_set] have "(v,t) \<in> supt" .
 with `(s,v) \<in> R` show "(s,t) \<in> supt O R" by auto
qed

lemmas rstep_seq_commute_supt = ctxt_closed_imp_commute_supt[OF ctxt_closed_rstep_seq]

lemma rstep_seq_commute_supteq: "(rstep R)^* O supteq \<subseteq> supteq O (rstep R)^*" (is "?S \<subseteq> ?T")
proof (rule subsetI2)
  fix s t assume "(s,t) \<in> ?S"
  hence "(s,t) \<in> (rstep R)^* O supt \<union> (rstep R)^*" by auto
  thus "(s,t) \<in> ?T"
  proof
    assume "(s,t) \<in> (rstep R)^* O supt" thus ?thesis using rstep_seq_commute_supt by auto
  next
    assume "(s,t) \<in> (rstep R)^*" thus ?thesis by auto
  qed
qed

lemma quasi_commute_supt_rstep_seq: "quasi_commute ((rstep R)^*) supt" unfolding quasi_commute_def
using rstep_seq_commute_supt[of R] by auto

lemma supt_sub_closed: "sub_closed supt" unfolding sub_closed_def
proof (rule subsetI2)
 fix s t assume "(s,t) \<in> CS supt"
 then obtain u v \<sigma> where s: "s = u \<cdot> \<sigma>" and t: "t = v \<cdot> \<sigma>" and "(u,v) \<in> supt" by (cases rule: closure_sub.cases) auto
 thus "(s,t) \<in> supt" unfolding s t using suptp_stable[to_set] by simp
qed

lemma supt_pi_sub_closed:
 fixes p :: "'f sp"
 defines "\<pi> \<equiv> lift2term p"
 assumes "\<pi> s \<rhd> \<pi> t" shows "(\<pi> s)\<cdot>\<sigma> \<rhd> (\<pi> t)\<cdot>\<sigma>" using assms suptp_stable by simp

lemma rstep_seq_pi_sub_closed:
 fixes p :: "'f sp"
 defines "\<pi> \<equiv> lift2term p"
 assumes "(\<pi> s,\<pi> t) \<in> (rstep R)^*" shows "((\<pi> s)\<cdot>\<sigma>,(\<pi> t)\<cdot>\<sigma>) \<in> (rstep R)^*" using assms rstep_seq_sub_closed unfolding sub_closed_def by blast

lemma lift2term_sub_distr[simp]:
  assumes "\<pi> f < length ss" shows "lift2term \<pi> (Fun f ss) \<cdot> \<sigma> = lift2term \<pi> (Fun f ss \<cdot> \<sigma>)" using assms by auto

lemma supt_neutral: "supt O supteq = supt"
proof
  show "supt O supteq \<subseteq> supt"
  proof (rule subsetI2)
    fix x y assume "(x,y) \<in> supt O supteq"
    then obtain z where "(x,z) \<in> supteq" and "(z,y) \<in> supt" by auto
    from `(x,z) \<in> supteq` show "(x,y) \<in> supt"
    proof (cases rule: supteq_cases)
      assume "x = z" with `(z,y) \<in> supt` show ?thesis by simp
    next
      assume "(x,z) \<in> supt" from this and `(z,y) \<in> supt` show ?thesis by (rule supt_trans)
    qed
  qed
next
  show "supt \<subseteq> supt O supteq" by (cases rule: supteq_cases) auto
qed

lemma nondef_root_imp_arg_step:
  assumes "(Fun f ss,t) \<in> rstep R"
      and wf: "\<forall>lr\<in>R. \<forall>x. fst lr \<noteq> Var x"
      and "f \<notin> DefSymbs R"
  shows "\<exists>i < length ss. (ss!i,t|_<i>) \<in> rstep R \<and> t = Fun f (take i ss@t|_<i>#drop (Suc i) ss)"
proof -
  from assms obtain l r p \<sigma> where rstep_r_p_s: "(Fun f ss,t) \<in> rstep_r_p_s R (l,r) p \<sigma>" unfolding rstep_iff_rstep_r_p_s by auto
  let ?C = "ctxt_of_pos_term p (Fun f ss)"
  from rstep_r_p_s have "p \<in> poss(Fun f ss)" and "(l,r) \<in> R"
    and "?C\<langle>l\<cdot>\<sigma>\<rangle> = Fun f ss" and "?C\<langle>r\<cdot>\<sigma>\<rangle> = t" unfolding rstep_r_p_s_def Let_def by auto
  have "\<exists>i q. p = i<#q"
  proof (cases p)
    case PCons thus ?thesis by auto
  next
    case Empty
    have "?C = \<box>" unfolding Empty by simp
    with `?C\<langle>l\<cdot>\<sigma>\<rangle> = Fun f ss` have "l\<cdot>\<sigma> = Fun f ss" by simp
    have "\<forall>(l,r)\<in>R. \<exists>f ss. l = Fun f ss"
    proof (intro ballI impI)
      fix lr assume "lr \<in> R"
      with wf have "\<forall>x. fst lr \<noteq> Var x" by auto
      hence "\<exists>f ss. (fst lr) = Fun f ss" by (cases "fst lr") auto
      thus "(\<lambda>(l,r). \<exists>f ss. l = Fun f ss) lr" by auto
    qed
    with `(l,r) \<in> R` obtain g ts where "l = Fun g ts" unfolding wf_trs_def by best
    with `l\<cdot>\<sigma> = Fun f ss` have "f = g" by auto
    from `l = Fun g ts` and `(l,r) \<in> R` have "f \<in> DefSymbs R" unfolding DefSymbs_def dsymb_def `f = g` by auto
    with `f \<notin> DefSymbs R` show ?thesis by simp
  qed
  then obtain i q where "p = i<#q" by auto
  from `p \<in> poss(Fun f ss)` have "i < length ss" and "q \<in> poss(ss!i)" unfolding `p = i<#q` by auto
  let ?D = "ctxt_of_pos_term q (ss!i)"
  have C: "?C = More f (take i ss) ?D (drop (Suc i) ss)" unfolding `p = i<#q` by auto
  from `?C\<langle>l\<cdot>\<sigma>\<rangle> = Fun f ss` have "take i ss@?D\<langle>l\<cdot>\<sigma>\<rangle>#drop (Suc i) ss = ss" unfolding C by auto
  hence "(take i ss@?D\<langle>l\<cdot>\<sigma>\<rangle>#drop (Suc i) ss)!i = ss!i" by simp
  with `i < length ss` have "?D\<langle>l\<cdot>\<sigma>\<rangle> = ss!i" using take_is_nth[where xs = ss and i = i] by auto
  have t: "t = Fun f (take i ss@?D\<langle>r\<cdot>\<sigma>\<rangle>#drop (Suc i) ss)" unfolding `?C\<langle>r\<cdot>\<sigma>\<rangle> = t`[symmetric] C by simp
  from `i < length ss` have "t|_<i> = ?D\<langle>r\<cdot>\<sigma>\<rangle>" unfolding t unfolding subt_at.simps using take_is_nth[where xs = ss and i = i] by auto
  from `q \<in> poss(ss!i)` and `(l,r) \<in> R` 
    and `?D\<langle>l\<cdot>\<sigma>\<rangle> = ss!i` and `t|_<i> = ?D\<langle>r\<cdot>\<sigma>\<rangle>`[symmetric]
    have "(ss!i,t|_<i>) \<in> rstep_r_p_s R (l,r) q \<sigma>" unfolding rstep_r_p_s_def Let_def by auto
  hence "(ss!i,t|_<i>) \<in> rstep R" unfolding rstep_iff_rstep_r_p_s by auto
  from `i < length ss` and this and t show ?thesis unfolding `t|_<i> = ?D\<langle>r\<cdot>\<sigma>\<rangle>`[symmetric] by auto
qed

lemma nondef_root_imp_arg_steps:
  assumes "(Fun f ss,t) \<in> (rstep R)^*"
      and wf: "\<forall>lr\<in>R. \<forall>x. fst lr \<noteq> Var x"
      and "f \<notin> DefSymbs R"
  shows "\<exists>ts. length ts = length ss \<and> t = Fun f ts \<and> (\<forall>i<length ss. (ss!i,ts!i) \<in> (rstep R)^*)"
proof -
  from assms obtain n where "(Fun f ss,t) \<in> (rstep R)^n" using rtrancl_imp_rel_pow' by best
  thus ?thesis
  proof (induct n arbitrary: t)
    case 0 thus ?case by auto
  next
    case (Suc n)
    then obtain u where "(Fun f ss,u) \<in> (rstep R)^n" and "(u,t) \<in> rstep R" by auto
    with Suc obtain ts where IH1: "length ts = length ss" and IH2: "u = Fun f ts" and IH3: "\<forall>i<length ss. (ss!i,ts!i) \<in> (rstep R)^*" by auto
    from `(u,t) \<in> rstep R` have "(Fun f ts,t) \<in> rstep R" unfolding `u = Fun f ts` .
    from nondef_root_imp_arg_step[OF this wf `f \<notin> DefSymbs R`]
      obtain j where "j < length ts"
      and "(ts!j,t|_<j>) \<in> rstep R"
      and B: "t = Fun f (take j ts@t|_<j>#drop (Suc j) ts)" (is "t = Fun f ?ts") by auto
    from `j < length ts` have "length ?ts = length ts" by auto
    hence A: "length ?ts = length ss" unfolding `length ts = length ss` .
    have C: "\<forall>i<length ss. (ss!i,?ts!i) \<in> (rstep R)^*"
    proof (intro allI, intro impI)
      fix i assume "i < length ss"
        from `i < length ss` and IH3 have "(ss!i,ts!i) \<in> (rstep R)^*" by auto
      have "i = j \<or> i \<noteq> j" by auto
      thus "(ss!i,?ts!i) \<in> (rstep R)^*"
      proof
        assume "i = j"
        from `j < length ts` have "j \<le> length ts" by simp
        from take_is_nth[OF this] have "?ts!j = t|_<j>" by simp
        from `(ts!j,t|_<j>) \<in> rstep R` have "(ts!i,t|_<i>) \<in> rstep R" unfolding `i = j` .
        with `(ss!i,ts!i) \<in> (rstep R)^*` show ?thesis unfolding `i = j` unfolding `?ts!j = t|_<j>` by auto
      next
        assume "i \<noteq> j"
        from `i < length ss` have "i \<le> length ts" unfolding `length ts = length ss` by simp
        from `j < length ts` have "j \<le> length ts" by simp
        from nth_take_drop_is_nth[OF `i \<le> length ts` `j \<le> length ts` `i \<noteq> j`]
          have "?ts!i = ts!i" by simp
        with `(ss!i,ts!i) \<in> (rstep R)^*` show ?thesis by auto
      qed
    qed
    from A and B and C show ?case by blast
  qed
qed

lemma rsteps_comp_supt_subset_supt_union_rstep_trancl:
  "(rstep R)^* O supt \<subseteq> (supt \<union> rstep R)^+"
proof (rule subsetI2)
  fix s t assume "(s,t) \<in> (rstep R)^* O supt"
  then obtain u where "(u,t) \<in> (rstep R)^*"
    and "(s,u) \<in> supt" by auto
  thus "(s,t) \<in> (supt \<union> rstep R)^+"
  proof (induct)
    case base thus ?case by auto
  next
    case (step v w) hence IH: "(s,v) \<in> (supt \<union> rstep R)^+" by auto
    from `(v,w) \<in> rstep R` have "(v,w) \<in> supt \<union> rstep R" by auto
    hence "(v,w) \<in> (supt \<union> rstep R)^+" by auto
    with IH show ?case by simp
  qed
qed

lemma subterm_preserves_SN:
  fixes R :: "('f,'v)trel" and t :: "('f,'v)term" and s :: "('f,'v)term"
  assumes "SN (rstep R) t" and "(t,s) \<in> supt"
  shows "SN (rstep R) s"
proof (rule ccontr)
  assume "\<not> SN (rstep R) s"
  then obtain S where "S 0 = s" and seq: "\<forall>i. (S i,S(Suc i)) \<in> rstep R" unfolding SN_ars_def SN_elt_def by auto
  from `(t,s) \<in> supt`[to_pred] obtain C where "t = C\<langle>s\<rangle>" using suptp_imp_nectxt by auto
  let ?S = "\<lambda>i. C\<langle>S i\<rangle>"
  from `S 0 = s` have "?S 0 = t" unfolding `t = C\<langle>s\<rangle>` by auto
  moreover have "\<forall>i. (?S i,?S(Suc i)) \<in> rstep R" using seq using rstep_imp_ctxt_rstep by blast
  ultimately have "\<not> SN (rstep R) t" unfolding SN_ars_def SN_elt_def by auto
  with assms show False by auto
qed

lemma SN_rstep_seq_imp_SN_supt_union_rstep_seq:
  fixes R :: "('f,'v)trel" and t :: "('f,'v)term"
  assumes "SN (rstep R) t"
  shows "SN (supt \<union> rstep R) t" (is "SN ?S t")
proof (rule ccontr)
  assume "\<not> SN ?S t"
  then obtain s where ini: "s 0 = t" and seq: "\<forall>i. (s i,s(Suc i)) \<in> ?S"
    unfolding SN_elt_def by auto
  have SN: "\<forall>i. SN (rstep R) (s i)"
  proof
    fix i show "SN (rstep R) (s i)"
    proof (induct i)
      case 0 show ?case using assms unfolding `s 0 = t` .
    next
      case (Suc i)
      from seq have "(s i,s(Suc i)) \<in> ?S" by simp
      thus ?case
      proof
        assume "(s i,s(Suc i)) \<in> supt"
        from subterm_preserves_SN[OF Suc this] show ?thesis .
      next
        assume "(s i,s(Suc i)) \<in> rstep R"
        from step_preserves_SN[OF this Suc] show ?thesis .
      qed
    qed
  qed
  have no_seq: "\<not>(\<exists>S. \<forall>i. (S i,S(Suc i)) \<in> snRelSupt(rstep R))"
    using SN_snRelRstepSupt unfolding SN_ars_def SN_elt_def by auto
  have "\<forall>i. (s i,s(Suc i)) \<in> snRelSupt(rstep R)"
  proof
    fix i
    from SN have SN: "SN (rstep R) (s i)" by simp
    from seq have "(s i,s(Suc i)) \<in> supt \<union> rstep R" by simp
    thus "(s i,s(Suc i)) \<in> snRelSupt(rstep R)"
    proof
      assume "(s i,s(Suc i)) \<in> supt"
      thus ?thesis unfolding snRelSupt_def snRel_def using SN by auto
    next
      assume "(s i,s(Suc i)) \<in> rstep R"
      thus ?thesis unfolding snRelSupt_def snRel_def using SN by auto
    qed
  qed
  with no_seq show False by auto
qed

text {* Lhss and rhss of @{text "P"} are nonvariable and nonconstant terms with roots that are
not defined in @{text "R"}. *}
definition spscCond :: "'f sp \<Rightarrow> ('f,'v)trel \<Rightarrow> ('f,'v)trel \<Rightarrow> ('f,'v)trel \<Rightarrow> bool" where
  "spscCond p P P' R \<equiv> (\<forall>(l,r)\<in>P-P'. (lift2term p l,lift2term p r) \<in> supt)
  \<and> (\<forall>(l,r)\<in>P.
      (lift2term p l,lift2term p r) \<in> supteq
    \<and> (\<exists>f ss. l = Fun f ss \<and> 0 < length ss \<and> f \<notin> DefSymbs R \<and> p f < length ss)
    \<and> (\<exists>f ss. r = Fun f ss \<and> 0 < length ss \<and> f \<notin> DefSymbs R \<and> p f < length ss))"

definition subterm_criterion_proc :: "'f sp \<Rightarrow> ('f,'v)trel \<Rightarrow> ('f,'v)trel \<Rightarrow> ('f,'v)trel \<Rightarrow> ('f,'v)trel \<Rightarrow> bool" where
  "subterm_criterion_proc p P R P' R' \<equiv> (\<forall>lr\<in>R. \<forall>x. fst lr \<noteq> Var x) \<and> spscCond p P P' R \<and> R = R'"

lemma ci_subterm_criterion_proc: "chain_identifying_proc(subterm_criterion_proc p)"
  unfolding chain_identifying_proc_def
proof (intro allI, intro impI)
  fix P R P' R' assume cond: "subterm_criterion_proc p P R P' R'"
  from cond have "R' = R" and wf: "\<forall>lr\<in>R. \<forall>x. fst lr \<noteq> Var x" unfolding subterm_criterion_proc_def by auto
  hence R_mono: "R' \<subseteq> R" by simp
  have min_ichain_to_ichain: "\<forall>s t \<sigma>. min_ichain (P,R) s t \<sigma> \<longrightarrow>
    (\<exists>i. ichain (P',R') (shift s i) (shift t i) (shift \<sigma> i))"
  proof (intro allI, intro impI)
    fix s t \<sigma> assume "min_ichain (P,R) s t \<sigma>"
    hence "ichain (P,R) s t \<sigma>" and min: "minimal_cond R s t \<sigma>" by auto
    hence dpsteps: "\<forall>i. (s i,t i) \<in> P"
      and rsteps: "\<forall>i. (t i \<cdot> \<sigma> i,s(Suc i) \<cdot> \<sigma>(Suc i)) \<in> (rstep R)^*" by auto
    from min have min_t: "\<forall>i. SN (rstep R) (t i \<cdot> \<sigma> i)" unfolding minimal_cond_def .
    have min_s: "\<forall>i>(0::nat). SN (rstep R) (s i \<cdot> \<sigma> i)"
    proof (intro allI, intro impI)
      fix i :: nat assume "i > 0"
      then obtain n where "i = Suc n" by (cases i) auto
      from rsteps have step: "(t(i - 1) \<cdot> \<sigma>(i - 1),s i \<cdot> \<sigma> i) \<in> (rstep R)^*" unfolding `i = Suc n` by auto
      from min_t have "SN (rstep R) (t(i - 1) \<cdot> \<sigma>(i - 1))" unfolding `i = Suc n` by auto
      from rtrancl_step_preserves_SN[OF step this] show "SN (rstep R) (s i \<cdot> \<sigma> i)" .
    qed
    have "(\<exists>i. \<forall>j\<ge>i. (s j,t j) \<in> P') \<or> (\<forall>i. \<exists>j\<ge>i. (s j,t j) \<notin> P')" by auto
    thus "\<exists>i. ichain (P',R') (shift s i) (shift t i) (shift \<sigma> i)"
    proof
      assume "\<exists>i. \<forall>j\<ge>i. (s j,t j) \<in> P'"
      then obtain i where tail: "\<forall>j\<ge>i. (s j,t j) \<in> P'" by auto
      let ?s = "shift s i"
      let ?t = "shift t i"
      let ?\<sigma> = "shift \<sigma> i"
      from tail have A: "\<forall>j. (?s j,?t j) \<in> P'" by auto
      from rsteps have "\<forall>j. (?t j \<cdot> ?\<sigma> j,?s(Suc j) \<cdot> ?\<sigma>(Suc j)) \<in> (rstep R)^*" by auto
      with A show ?thesis unfolding ichain.simps unfolding `R' = R` by auto
    next
      let ?\<pi> = "lift2term p"
      let ?f = "\<lambda>i. ?\<pi>(s(Suc i)) \<cdot> \<sigma>(Suc i)"
      let ?g = "\<lambda>i. ?\<pi>(t(Suc i)) \<cdot> \<sigma>(Suc i)"
      let ?NS = "(rstep R)^*"
      let ?S = "(supt \<union> rstep R)^+"
      assume "\<forall>i. \<exists>j\<ge>i. (s j,t j) \<notin> P'"
      with dpsteps have strict: "\<forall>i. \<exists>j\<ge>Suc i. (s j,t j) \<in> P - P'" by auto
      from cond have "\<forall>(l,r)\<in>P-P'. (?\<pi>(l),?\<pi>(r)) \<in> supt" unfolding subterm_criterion_proc_def spscCond_def by auto
      with strict have "\<forall>i. \<exists>j\<ge>Suc i. (?\<pi>(s j),?\<pi>(t j)) \<in> supt" by blast
      hence "\<forall>i. \<exists>j\<ge>Suc i. (?\<pi>(s j) \<cdot> \<sigma> j,?\<pi>(t j) \<cdot> \<sigma> j) \<in> supt" using suptp_stable[to_set] by fast
      hence neq: "\<forall>i. \<exists>j\<ge>Suc i. ?\<pi>(s j) \<cdot> \<sigma> j \<noteq> ?\<pi>(t j) \<cdot> \<sigma> j" using suptp_imp_neq[to_set] by auto
      have neq_Suc: "\<forall>i. \<exists>j\<ge>i. ?f j \<noteq> ?g j"
      proof (intro allI)
        fix i from neq have "\<exists>j\<ge>Suc i. ?\<pi>(s j) \<cdot> \<sigma> j \<noteq> ?\<pi>(t j) \<cdot> \<sigma> j" by auto
        then obtain j where "j \<ge> Suc i" and ne: "?\<pi>(s j) \<cdot> \<sigma> j \<noteq> ?\<pi>(t j) \<cdot> \<sigma> j" by auto
        hence "j > i" by auto
        from less_imp_Suc_add[OF this]
          have "\<exists>m. j = Suc(i + m)" by simp
        then obtain m where "j = Suc(i + m)" by auto
        hence "Suc(i + m) > i" using `j > i` by simp
        hence "i + m \<ge> i" by auto
        from ne have "?f(i + m) \<noteq> ?g(i + m)" unfolding `j = Suc(i+m)` by simp
        with `i + m \<ge> i` have "i + m \<ge> i \<and> ?f(i + m) \<noteq> ?g(i + m)" by auto
        thus "\<exists>j\<ge>i. ?f j \<noteq> ?g j" by (rule exI)
      qed
        

      have "\<forall>i. (?\<pi>(s i),?\<pi>(t i)) \<in> supteq"
      proof
        fix i
        from dpsteps and cond show "(?\<pi>(s i),?\<pi>(t i)) \<in> supteq"  unfolding subterm_criterion_proc_def spscCond_def by auto
      qed
      hence supteq: "\<forall>i. (?\<pi>(s i)\<cdot>\<sigma> i,?\<pi>(t i)\<cdot>\<sigma> i) \<in> supteq" using supteqp_stable[to_set] by auto

      have proj: "\<forall>i. (?\<pi>(t i) \<cdot> \<sigma> i,?\<pi>(s(Suc i)) \<cdot> \<sigma>(Suc i)) \<in> (rstep R)^*"
      proof
        fix i
        from rsteps have steps: "(t i \<cdot> \<sigma> i,s(Suc i) \<cdot> \<sigma>(Suc i)) \<in> (rstep R)^*" by simp
        from dpsteps and cond have "\<exists>f ss. t i = Fun f ss \<and> 0 < length ss \<and> f \<notin> DefSymbs R \<and> p f < length ss" unfolding subterm_criterion_proc_def spscCond_def by blast
        then obtain f ss where "t i = Fun f ss"
          and "0 < length ss"
          and "f \<notin> DefSymbs R"
          and "p f < length ss" by auto
        from dpsteps and cond
          have "\<exists>f ss. s(Suc i) = Fun f ss \<and> 0 < length ss \<and> f \<notin> DefSymbs R \<and> p f < length ss" unfolding subterm_criterion_proc_def spscCond_def by blast
        then obtain g ts where "s(Suc i) = Fun g ts"
          and "0 < length ts"
          and "g \<notin> DefSymbs R"
          and "p g < length ts" by auto
        from steps
          have "(Fun f (map (\<lambda>t. t \<cdot> \<sigma> i) ss),s(Suc i) \<cdot> \<sigma>(Suc i)) \<in> (rstep R)^*"
          (is "(Fun f ?ss,?s) \<in> (rstep R)^*")
          unfolding `t i = Fun f ss` by simp
        from nondef_root_imp_arg_steps[OF this wf `f \<notin> DefSymbs R`]
          obtain us where "length us = length ?ss"
          and "?s = Fun f us"
          and args: "\<forall>j<length ?ss. (?ss!j,us!j) \<in> (rstep R)^*" by auto
        have "?s = Fun g (map (\<lambda>t. t \<cdot> \<sigma>(Suc i)) ts)"
          (is "_ = Fun g ?ts") unfolding `s(Suc i) = Fun g ts` by simp
        from `?s = Fun f us` and `?s = Fun g ?ts`
          have "f = g" and "us = ?ts" by auto
        from `p f < length ss` have "p f < length ?ss" by auto
        have "?\<pi>(t i) \<cdot> \<sigma> i = ss!(p f) \<cdot> \<sigma> i" unfolding `t i = Fun f ss` by simp
        hence t: "?\<pi>(t i) \<cdot> \<sigma> i = ?ss!(p f)" unfolding nth_map[OF `p f < length ss`] .
        from `p g < length ts` have "p f < length ts" unfolding `f = g` .
        have "?\<pi>(s(Suc i)) \<cdot> \<sigma>(Suc i) = ts!(p f) \<cdot> \<sigma>(Suc i)" unfolding `f = g` `s(Suc i) = Fun g ts` by auto
        hence "?\<pi>(s(Suc i)) \<cdot> \<sigma>(Suc i) = ?ts!(p f)" unfolding nth_map[OF `p f < length ts`] .
        hence s: "?\<pi>(s(Suc i)) \<cdot> \<sigma>(Suc i) = us!(p f)" unfolding `us = ?ts` .
        from `p f < length ?ss` args show "(?\<pi>(t i) \<cdot> \<sigma> i,?\<pi>(s(Suc i)) \<cdot> \<sigma>(Suc i)) \<in> (rstep R)^*" unfolding s t by auto
      qed
      

      have "\<forall>i. (?f i,?f(Suc i)) \<in> ?NS \<union> ?S"
      proof
        fix i
        from supteq and proj have "(?f i,?f(Suc i)) \<in> (rstep R)^* O supteq" by auto
        then obtain u where "(?f i,u) \<in> supteq" and "(u,?f(Suc i)) \<in> (rstep R)^*" by auto
        hence "?f i = u \<or> (?f i,u) \<in> supt" by auto
        thus "(?f i,?f(Suc i)) \<in> ?NS \<union> ?S"
        proof
          assume "?f i = u"
          with `(u,?f(Suc i)) \<in> (rstep R)^*` show ?thesis by auto
        next
          assume "(?f i,u) \<in> supt"
          with `(u,?f(Suc i)) \<in> (rstep R)^*`
            have "(?f i,?f(Suc i)) \<in> (rstep R)^* O supt" by auto
          thus ?thesis using rsteps_comp_supt_subset_supt_union_rstep_trancl by auto
        qed
      qed

      moreover have "?S O ?NS \<subseteq> ?S"
      proof (rule subsetI2)
        fix s t assume "(s,t) \<in> ?S O ?NS"
        then obtain u where "(s,u) \<in> ?NS" and "(u,t) \<in> ?S" by auto
        thus "(s,t) \<in> ?S"
        proof (induct)
          case base thus ?case by simp
        next
          case (step v w)
          from `(v,w) \<in> rstep R` have "(v,w) \<in> ?S" by auto
          with `(w,t) \<in> ?S` have "(v,t) \<in> ?S" by auto
          with step show ?case by auto
        qed
      qed

      moreover have "SN ?S (?f 0)"
      proof -
        have "\<forall>i. (s i,?\<pi>(s i)) \<in> supt"
        proof
          fix i
          from dpsteps and cond have "\<exists>f ss. s i = Fun f ss \<and> 0 < length ss \<and> f \<notin> DefSymbs R \<and> p f < length ss" unfolding subterm_criterion_proc_def spscCond_def by blast
          then obtain f ss where "s i = Fun f ss" and "0 < length ss" and "p f < length ss" by best
          have "?\<pi>(s i) = ss!(p f)" unfolding `s i = Fun f ss` by simp
          from `p f < length ss` have "ss!(p f) \<in> set ss" by simp
          from supt.arg[OF this] show "(s i,?\<pi>(s i)) \<in> supt" unfolding `?\<pi>(s i) = ss!(p f)` `s i = Fun f ss` by simp
        qed
        with suptp_stable[to_set]
          have "\<forall>i. (s i \<cdot> \<sigma> i,?\<pi>(s i) \<cdot> \<sigma> i) \<in> supt" by auto
        hence supt: "(s(Suc 0) \<cdot> \<sigma>(Suc 0),?f 0) \<in> supt" by auto
        from `\<forall>i>0. SN (rstep R) (s i \<cdot> \<sigma> i)`
          have "SN (rstep R) (s(Suc 0) \<cdot> \<sigma>(Suc 0))" by auto
        from subterm_preserves_SN[OF this supt]
          have "SN (rstep R) (?f 0)" by auto
        from SN_rstep_seq_imp_SN_supt_union_rstep_seq[OF `SN (rstep R) (?f 0)`]
          have "SN (supt \<union> rstep R) (?f 0)" .
        from SN_step_imp_SN_trancl[OF this] show "SN ?S (?f 0)" .
      qed

      ultimately have "\<exists>j. \<forall>i\<ge>j. (?f i,?f(Suc i)) \<in> ?NS - ?S" by (rule non_strict_ending)
      then obtain j where nonstrict: "\<forall>i\<ge>j. (?f i,?f(Suc i)) \<in> ?NS - ?S" by auto
      {
        fix i 
        assume "i \<ge> j" have "?f i = ?g i"
        proof (rule ccontr)
          assume "?f i \<noteq> ?g i"
          with supteq have "(?f i,?g i) \<in> supt" by auto
          with proj have "(?f i,?f(Suc i)) \<in> (rstep R)^* O supt" by auto
          hence "(?f i,?f(Suc i)) \<in> ?S" using rsteps_comp_supt_subset_supt_union_rstep_trancl by auto
          with nonstrict and `i \<ge> j` show "False" by auto
        qed
      }
      hence "\<forall>i\<ge>j. ?f i = ?g i" by auto
      with neq_Suc have False by auto
      thus ?thesis by simp
    qed
  qed
  with R_mono show "R' \<le> R \<and> (\<forall>s t \<sigma>. min_ichain (P,R) s t \<sigma> \<longrightarrow>
    (\<exists>i. ichain (P',R') (shift s i) (shift t i) (shift \<sigma> i)))" by auto
qed

end
